@@ -98,25 +98,29 @@ enum
   DEE_MODEL_SIGNAL_ROW_ADDED,
   DEE_MODEL_SIGNAL_ROW_REMOVED,
   DEE_MODEL_SIGNAL_ROW_CHANGED,
+  DEE_MODEL_SIGNAL_CHANGESET_STARTED,
+  DEE_MODEL_SIGNAL_CHANGESET_FINISHED,
 
   DEE_MODEL_LAST_SIGNAL
 };
 
 static guint32 dee_model_signals[DEE_MODEL_LAST_SIGNAL] = { 0 };
 
-#define CHECK_SCHEMA(self,return_expression) \
-if (G_UNLIKELY (dee_model_get_schema (self, NULL) == NULL)) \
+#define CHECK_SCHEMA(self,out_num_cols,return_expression) \
+if (G_UNLIKELY (dee_model_get_schema (self, out_num_cols) == NULL)) \
   { \
     g_critical ("The model %s@%p doesn't have a schema", \
                 G_OBJECT_TYPE_NAME (self), self); \
     return_expression; \
   }
 
-#define DEFAULT_NUM_COLUMNS 16
-
 static void            dee_model_set_schema_valist (DeeModel    *self,
                                                     va_list     *args);
 
+static void            dee_model_set_column_names_valist (DeeModel    *self,
+                                                          const gchar *first_column_name,
+                                                          va_list     *args);
+
 static DeeModelIter*   dee_model_append_valist  (DeeModel *self,
                                                  va_list  *args);
 
@@ -227,6 +231,44 @@ dee_model_default_init (DeeModelInterface *klass)
                   g_cclosure_marshal_VOID__BOXED,
                   G_TYPE_NONE, 1,
                   DEE_TYPE_MODEL_ITER);
+
+  /**
+   * DeeModel::changeset-started
+   * @self: the #DeeModel on which the signal is emitted
+   *
+   * Connect to this signal to be notified when a changeset that can contain
+   * multiple row additions / changes / removals is about to be committed
+   * to the model.
+   * Note that not all model implementations use the changeset approach and
+   * you might still get a row change signal outside of changeset-started and
+   * changeset-finished signals. It also isn't guaranteed that a changeset
+   * would always be non-empty.
+   */
+  dee_model_signals[DEE_MODEL_SIGNAL_CHANGESET_STARTED] =
+    g_signal_new ("changeset-started",
+                  DEE_TYPE_MODEL,
+                  G_SIGNAL_RUN_LAST,
+                  G_STRUCT_OFFSET (DeeModelIface, changeset_started),
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__VOID,
+                  G_TYPE_NONE, 0);
+
+  /**
+   * DeeModel::changeset-finished
+   * @self: the #DeeModel on which the signal is emitted
+   *
+   * Connect to this signal to be notified when a changeset that can contain
+   * multiple row additions / changes / removals has been committed
+   * to the model.
+   */
+  dee_model_signals[DEE_MODEL_SIGNAL_CHANGESET_FINISHED] =
+    g_signal_new ("changeset-finished",
+                  DEE_TYPE_MODEL,
+                  G_SIGNAL_RUN_LAST,
+                  G_STRUCT_OFFSET (DeeModelIface, changeset_finished),
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__VOID,
+                  G_TYPE_NONE, 0);
 }
 
 /**
@@ -349,6 +391,107 @@ dee_model_set_schema_full (DeeModel           *self,
   (* iface->set_schema_full) (self, column_schemas, num_columns);
 }
 
+void
+dee_model_set_column_names (DeeModel *self, const gchar *first_column_name,
+                            ...)
+{
+  va_list        args;
+
+  g_return_if_fail (DEE_IS_MODEL (self));
+
+  va_start (args, first_column_name);
+  dee_model_set_column_names_valist (self, first_column_name, &args);
+  va_end (args);
+}
+
+static void
+dee_model_set_column_names_valist (DeeModel *self,
+                                   const gchar *first_column_name,
+                                   va_list *args)
+{
+  DeeModelIface *iface;
+  const gchar  **column_names;
+  guint          n_columns, i;
+
+  g_return_if_fail (DEE_IS_MODEL (self));
+
+  n_columns = dee_model_get_n_columns (self);
+
+  g_return_if_fail (n_columns != 0);
+
+  column_names = g_alloca (sizeof (gchar*) * n_columns);
+  column_names[0] = first_column_name;
+
+  i = 1;
+  /* Extract and validate the column schema strings from the va_list */
+  while (i < n_columns)
+    {
+      gchar *name = va_arg (*args, gchar*);
+      column_names[i++] = name;
+      if (name == NULL) break;
+    }
+
+  iface = DEE_MODEL_GET_IFACE (self);
+  (* iface->set_column_names_full) (self, column_names, i);
+}
+
+/**
+ * dee_model_set_column_names_full:
+ * @self: A #DeeModel.
+ * @column_names: (array length=num_columns zero-terminated=1) (element-type utf8) (transfer none): A list of column names terminated by a %NULL
+ * @num_columns: an integer specifying the array length for @annotations
+ *
+ * Set column names used by @self.
+ * This method must be called exactly once, but only after setting
+ * a schema of the model. Note that some constructors will do this for you.
+ */
+void
+dee_model_set_column_names_full (DeeModel     *self,
+                                 const gchar **column_names,
+                                 guint         num_columns)
+{
+  DeeModelIface *iface;
+
+  g_return_if_fail (DEE_IS_MODEL (self));
+
+  if (dee_model_get_schema (self, NULL) == NULL)
+    {
+      g_critical ("The model %s@%p doesn't have a schema set",
+                  G_OBJECT_TYPE_NAME (self), self);
+      return;
+    }
+
+  iface = DEE_MODEL_GET_IFACE (self);
+  (* iface->set_column_names_full) (self, column_names, num_columns);
+}
+
+/**
+ * dee_model_get_column_names:
+ * @self: The #DeeModel to get the the schema for
+ * @num_columns: (out) (allow-none): Address of an integer in which to store the
+ *               number of columns in @self. Or %NULL to ignore the array length.
+ *
+ * Get a %NULL-terminated array of column names for the columns of @self.
+ * These names can be used in calls to dee_model_build_named_row().
+ *
+ * Returns: (array length=num_columns) (element-type utf8) (transfer none):
+ *          A %NULL-terminated array of #GVariant type strings. The length of
+ *          the returned array is written to @num_columns. The returned array
+ *          should not be freed or modified. It is owned by the model.
+ */
+const gchar**
+dee_model_get_column_names (DeeModel *self,
+                            guint    *num_columns)
+{
+  DeeModelIface *iface;
+
+  g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
+
+  iface = DEE_MODEL_GET_IFACE (self);
+
+  return (* iface->get_column_names) (self, num_columns);
+}
+
 /**
  * dee_model_get_schema:
  * @self: The #DeeModel to get the the schema for
@@ -398,6 +541,111 @@ dee_model_get_column_schema (DeeModel *self,
   return (* iface->get_column_schema) (self, column);
 }
 
+/**
+ * dee_model_get_field_schema:
+ * @self: a #DeeModel
+ * @field_name: name of vardict field to get schema of
+ * @out_column: (out): column index of the associated vardict
+ *
+ * Get the #GVariant signature of field previously registered with 
+ * dee_model_register_vardict_schema().
+ *
+ * Return value: the #GVariant signature for the field, or %NULL if given field
+ *               wasn't registered with dee_model_register_vardict_schema().
+ */
+const gchar*
+dee_model_get_field_schema (DeeModel    *self,
+                            const gchar *field_name,
+                            guint       *out_column)
+{
+  DeeModelIface *iface;
+
+  g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
+
+  iface = DEE_MODEL_GET_IFACE (self);
+
+  return (* iface->get_field_schema) (self, field_name, out_column);
+}
+
+/**
+ * dee_model_get_column_index:
+ * @self: a #DeeModel
+ * @column_name: the column name to retrieve the index of
+ *
+ * Get the column index of a column.
+ *
+ * Return value: 0-based index of the column or -1 if column with this name 
+ *               wasn't found
+ */
+gint
+dee_model_get_column_index (DeeModel    *self,
+                            const gchar *column_name)
+{
+  DeeModelIface *iface;
+
+  g_return_val_if_fail (DEE_IS_MODEL (self), -1);
+
+  iface = DEE_MODEL_GET_IFACE (self);
+
+  return (* iface->get_column_index) (self, column_name);
+}
+
+/**
+ * dee_model_register_vardict_schema:
+ * @self: a #DeeModel
+ * @column: the column index to register the schemas with
+ * @schemas: (element-type utf8 utf8): hashtable with keys specifying
+ *           names of the fields and values defining their schema
+ *
+ * Register schema for fields in a model containing column with variant
+ * dictionary schema ('a{sv}').
+ * The keys registered with this function can be later used
+ * with dee_model_build_named_row() function, as well as
+ * dee_model_get_value_by_name(). Note that it is possible to register
+ * the same field name for multiple columns, in which case you need to use
+ * fully-qualified "column_name::field" name in the calls to
+ * dee_model_build_named_row() and dee_model_get_field_schema().
+ */
+void
+dee_model_register_vardict_schema (DeeModel   *self,
+                                   guint       column,
+                                   GHashTable *schemas)
+{
+  DeeModelIface *iface;
+
+  g_return_if_fail (DEE_IS_MODEL (self));
+
+  iface = DEE_MODEL_GET_IFACE (self);
+
+  return (* iface->register_vardict_schema) (self, column, schemas);
+}
+
+/**
+ * dee_model_get_vardict_schema:
+ * @self: a #DeeModel
+ * @column: the column index to get the schemas for
+ *
+ * Get a schema for variant dictionary column previously registered using
+ * dee_model_register_vardict_schema().
+ *
+ * Return value: (transfer container) (element-type utf8 utf8): Hashtable 
+ *               containing a mapping from field names to schemas or NULL.
+ *               Note that keys and values in the hashtable may be owned
+ *               by the model, so you need to create a deep copy if you
+ *               intend to keep the hashtable around.
+ */
+GHashTable*
+dee_model_get_vardict_schema (DeeModel *self, guint column)
+{
+  DeeModelIface *iface;
+
+  g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
+
+  iface = DEE_MODEL_GET_IFACE (self);
+
+  return (* iface->get_vardict_schema) (self, column);
+}
+
 /**
  * dee_model_get_n_columns:
  * @self: a #DeeModel
@@ -438,6 +686,103 @@ dee_model_get_n_rows (DeeModel *self)
   return (* iface->get_n_rows) (self);
 }
 
+/**
+ * dee_model_begin_changeset:
+ * @self: a #DeeModel
+ *
+ * Notify listeners that the model is about to be changed, which means that
+ * multiple row additions / changes / removals will follow.
+ * The default implementation of this method will emit
+ * the ::changeset-started signal.
+ *
+ * It is not stricly necessary to enclose every change to a model 
+ * in a dee_model_begin_changeset() and dee_model_end_changeset() calls, but
+ * doing so is highly recommended and allows implementing various optimizations.
+ *
+ * The usual way to perform multiple changes to a model is as follows:
+ *
+ * <programlisting>
+ * void update_model (DeeModel *model)
+ * {
+ *   GVariant **added_row_data1 = ...;
+ *   GVariant **added_row_data2 = ...;
+ *
+ *   dee_model_begin_changeset (model);
+ *
+ *   dee_model_remove (model, dee_model_get_first_iter (model));
+ *   dee_model_append_row (model, added_row_data1);
+ *   dee_model_append_row (model, added_row_data2);
+ *
+ *   dee_model_end_changeset (model);
+ * }
+ * </programlisting>
+ */
+void
+dee_model_begin_changeset (DeeModel *self)
+{
+  DeeModelIface *iface;
+
+  g_return_if_fail (DEE_IS_MODEL (self));
+
+  iface = DEE_MODEL_GET_IFACE (self);
+
+  if (iface->begin_changeset)
+    (* iface->begin_changeset) (self);
+  else
+    g_signal_emit (self, dee_model_signals[DEE_MODEL_SIGNAL_CHANGESET_STARTED], 0);
+}
+
+/**
+ * dee_model_end_changeset:
+ * @self: a #DeeModel
+ *
+ * Notify listeners that all changes have been committed to the model.
+ * The default implementation of this method will emit
+ * the ::changeset-finished signal.
+ *
+ * See also dee_model_begin_changeset().
+ */
+void
+dee_model_end_changeset (DeeModel *self)
+{
+  DeeModelIface *iface;
+
+  g_return_if_fail (DEE_IS_MODEL (self));
+
+  iface = DEE_MODEL_GET_IFACE (self);
+
+  if (iface->end_changeset)
+    (* iface->end_changeset) (self);
+  else
+    g_signal_emit (self, dee_model_signals[DEE_MODEL_SIGNAL_CHANGESET_FINISHED], 0);
+}
+
+static GVariant*
+collect_variant (const gchar* col_schema, va_list *args)
+{
+  const gchar *col_string;
+  GVariant    *result;
+
+  if (g_variant_type_is_basic (G_VARIANT_TYPE (col_schema)))
+    {
+      switch (col_schema[0])
+        {
+          case 's':
+          case 'o':
+          case 'g':
+            col_string = va_arg (*args, const gchar*);
+            result = g_variant_new (col_schema, col_string ? col_string : "");
+            break;
+          default:
+            result = g_variant_new_va (col_schema, NULL, args);
+        }
+    }
+  else
+    result = va_arg (*args, GVariant*);
+
+  return result;
+}
+
 /**
  * dee_model_build_row:
  * @self: The model to create a row for
@@ -460,7 +805,7 @@ dee_model_get_n_rows (DeeModel *self)
  *
  * Returns: If @out_row_members is %NULL a newly allocated array of variants
  *          will be returned and the array must be freed with g_free().
- *          If @out_row_members is non%NULL it will be reused, and variants in
+ *          If @out_row_members is non-%NULL it will be reused, and variants in
  *          the array may or may not have floating references, which means the
  *          caller must make sure that g_variant_ref_sink() and
  *          g_variant_unref() are called on them.
@@ -476,7 +821,7 @@ dee_model_build_row (DeeModel  *self,
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
 
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, NULL, return NULL);
 
   va_start (args, out_row_members);
   result = dee_model_build_row_valist (self, out_row_members, &args);
@@ -503,7 +848,6 @@ dee_model_build_row_valist (DeeModel  *self,
 {
   guint         i, n_cols;
   const gchar  *col_schema;
-  const gchar  *col_string;
   const gchar  *const *schema;
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
@@ -517,26 +861,8 @@ dee_model_build_row_valist (DeeModel  *self,
     {
       col_schema = schema[i];
 
-      /* Basic types are passed raw, while everything else is boxed in
-       * a GVariant already */
-      if (g_variant_type_is_basic (G_VARIANT_TYPE (col_schema)))
-        {
-          switch (col_schema[0])
-            {
-              case 's':
-              case 'o':
-              case 'g':
-                col_string = va_arg (*args, const gchar*);
-                out_row_members[i] = g_variant_new (col_schema,
-                                                    col_string ? col_string : "");
-                break;
-              default:
-                out_row_members[i] = g_variant_new_va (col_schema, NULL, args);
-            }
-        }
-      else
-        out_row_members[i] = va_arg (*args, GVariant*);
-        
+      out_row_members[i] = collect_variant (col_schema, args);
+
       if (G_UNLIKELY (out_row_members[i] == NULL))
         {
           g_critical ("Trying to build a row with a NULL member on position"
@@ -549,6 +875,299 @@ dee_model_build_row_valist (DeeModel  *self,
   return out_row_members;
 }
 
+/**
+ * dee_model_build_named_row:
+ * @self: The model to create a row for
+ * @out_row_members: An array to write the values to or %NULL to allocate
+ *                   a new array. If non-%NULL it must have a length
+ *                   that is longer or equal to the number of columns in @self
+ * @first_column_name: A column name
+ * @VarArgs: Value for given column, followed by more name/value pairs,
+ *           followed by %NULL. The passed names have to match the column names
+ *           (or field names registered with
+ *           dee_model_register_vardict_schema()) and values have to be set
+ *           according to schema of the given column or field.
+ *           Basic variant types are passed directly while any other types
+ *           must be boxed in a #GVariant, similar to dee_model_build_row().
+ *
+ * Build an array of #GVariant<!-- -->s with values from the variadic argument
+ * list according to the column names and model schema for @self.
+ * The caller must call g_variant_ref_sink() and g_variant_unref() 
+ * on all the returned variants and g_free() the array itself if %NULL
+ * was passed as @out_row_members.
+ *
+ * This is utility function and will not touch or modify @self in any way.
+ *
+ * For example, to append a row to model with signature ("s", "u", "s") and
+ * column names set to ("uri", "count", "description") you could do:
+ * <informalexample><programlisting>
+ *  GVariant    *row_buf[3];
+ *
+ *  dee_model_append_row (model,
+ *    dee_model_build_named_row (model, row_buf,
+ *                               "uri", "http://example.org",
+ *                               "count", 435,
+ *                               "description", "Example.org site", NULL));
+ * </programlisting></informalexample>
+ * 
+ * Returns: If @out_row_members is %NULL a newly allocated array of variants
+ *          will be returned and the array must be freed with g_free().
+ *          If @out_row_members is non-%NULL it will be reused, and variants in
+ *          the array may or may not have floating references, which means the
+ *          caller must make sure that g_variant_ref_sink() and
+ *          g_variant_unref() are called on them.
+ *
+ */
+GVariant**
+dee_model_build_named_row (DeeModel    *self,
+                           GVariant   **out_row_members,
+                           const gchar *first_column_name,
+                           ...)
+{
+  va_list    args;
+  GVariant **result;
+
+  g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
+
+  CHECK_SCHEMA (self, NULL, return NULL);
+
+  va_start (args, first_column_name);
+  result = dee_model_build_named_row_valist (self, out_row_members,
+                                             first_column_name, &args);
+  va_end (args);
+
+  return result;
+}
+
+/**
+ * dee_model_build_named_row_sunk:
+ * @self: The model to create a row for
+ * @out_row_members: (array): An array to write the values to or %NULL to
+ *                   allocate a new array. If non-%NULL it must have a length
+ *                   that is longer or equal to the number of columns in @self
+ * @out_array_length: (out): Length of the returned variant array
+ * @first_column_name: A column name
+ * @VarArgs: Value for given column, followed by more name/value pairs,
+ *           followed by %NULL. The passed names have to match the column names
+ *           and values have to be set according to schema of @self.
+ *           Basic variant types are passed directly while any other types
+ *           must be boxed in a #GVariant, similar to dee_model_build_row().
+ *
+ * Version of dee_model_build_named_row() for language bindings - as opposed to
+ * dee_model_build_named_row(), the returned variants will be strong
+ * references, therefore you always have to call g_variant_unref() on the items
+ * and g_free() the array itself if %NULL was passed as @out_row_members.
+ *
+ * If @out_row_members is non-%NULL, g_variant_unref() will be called
+ * on its elements (if also non-%NULL), which allows easy reuse of the array
+ * memory in loops.
+ *
+ * This is utility function and will not touch or modify @self in any way.
+ *
+ * Example of memory management for model with schema ("s", "i") and 
+ * column names ("uri", "count"):
+ * <informalexample><programlisting>
+ *  GVariant    **row_buf;
+ *
+ *  row_buf = dee_model_build_named_row_sunk (model, NULL, "uri", "file:///",
+ *                                            "count", 0, NULL);
+ *  dee_model_append_row (model, row_buf);
+ *
+ *  for (int i = 1; i < 100; i++)
+ *  {
+ *    dee_model_append_row (model,
+ *      dee_model_build_named_row_sunk (model, row_buf, "uri", "file:///",
+ *                                      "count", i, NULL));
+ *  }
+ *  
+ *  g_variant_unref (row_buf[0]);
+ *  g_variant_unref (row_buf[1]);
+ *  g_free (row_buf);
+ * </programlisting></informalexample>
+ * 
+ * Returns: (array length=out_array_length): If @out_row_members is %NULL
+ *          a newly allocated array of variants will be returned and the array
+ *          must be freed with g_free().
+ *          If @out_row_members is non-%NULL it will be reused. Variants in
+ *          the array will have strong references, which means the
+ *          caller must make sure that g_variant_unref() is called on them.
+ *
+ */
+GVariant**
+dee_model_build_named_row_sunk (DeeModel    *self,
+                                GVariant   **out_row_members,
+                                guint       *out_array_length,
+                                const gchar *first_column_name,
+                                ...)
+{
+  va_list    args;
+  guint      num_columns, i;
+  GVariant **result;
+
+  g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
+
+  CHECK_SCHEMA (self, &num_columns, return NULL);
+
+  if (out_row_members)
+    {
+      for (i = 0; i < num_columns; i++)
+        {
+          if (out_row_members[i]) g_variant_unref (out_row_members[i]);
+        }
+    }
+
+  va_start (args, first_column_name);
+  result = dee_model_build_named_row_valist (self, out_row_members,
+                                             first_column_name, &args);
+  va_end (args);
+
+  if (result)
+    {
+      for (i = 0; i < num_columns; i++)
+        {
+          g_variant_ref_sink (result[i]);
+        }
+    }
+
+  if (out_array_length)
+    *out_array_length = result != NULL ? num_columns : 0;
+
+  return result;
+}
+
+GVariant**
+dee_model_build_named_row_valist (DeeModel    *self,
+                                  GVariant   **out_row_members,
+                                  const gchar *first_column_name,
+                                  va_list     *args)
+{
+  DeeModelIface    *iface;
+  guint             n_cols, i;
+  gint              col_idx, last_unset_col;
+  gboolean         *variant_set;
+  GVariantBuilder **builders;
+  const gchar      *col_name, *col_schema;
+  const gchar      *const *schema;
+
+  g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
+
+  schema = dee_model_get_schema (self, &n_cols);
+
+  if (out_row_members == NULL)
+    out_row_members = g_new0 (GVariant*, n_cols);
+
+  variant_set = g_alloca (n_cols * sizeof (gboolean));
+  memset (variant_set, 0, n_cols * sizeof (gboolean));
+
+  builders = g_alloca (n_cols * sizeof (GVariantBuilder*));
+  memset (builders, 0, n_cols * sizeof (GVariantBuilder*));
+
+  iface = DEE_MODEL_GET_IFACE (self);
+
+  col_name = first_column_name;
+  while (col_name != NULL)
+    {
+      col_idx = (* iface->get_column_index) (self, col_name);
+      if (col_idx >= 0)
+        {
+          col_schema = schema[col_idx];
+
+          out_row_members[col_idx] = collect_variant (col_schema, args);
+
+          if (G_UNLIKELY (out_row_members[col_idx] == NULL))
+            {
+              g_critical ("Trying to build a row with a NULL member for column"
+                          " %s. This is probably an error in an application using"
+                          " libdee", col_name);
+              break;
+            }
+          else
+            {
+              variant_set[col_idx] = TRUE;
+            }
+        }
+      else
+        {
+          // check if we have hints
+          col_schema = (* iface->get_field_schema) (self, col_name, (guint*) &col_idx);
+          if (col_schema != NULL)
+            {
+              const gchar *key_name;
+              if (builders[col_idx] == NULL)
+                {
+                  builders[col_idx] = g_variant_builder_new (G_VARIANT_TYPE (schema[col_idx]));
+                }
+
+              key_name = strstr (col_name, "::");
+              key_name = key_name != NULL ? key_name + 2 : col_name;
+              g_variant_builder_add (builders[col_idx], "{sv}",
+                                     key_name,
+                                     collect_variant (col_schema, args));
+            }
+          else
+            {
+              g_warning ("Unable to find column index for \"%s\"", col_name);
+              /* need to break, cause there's no way to know size of the value */
+              break;
+            }
+        }
+      col_name = va_arg (*args, const gchar*);
+    }
+
+  /* Finish builders */
+  for (i = 0; i < n_cols; i++)
+    {
+      if (builders[i])
+        {
+          out_row_members[i] = g_variant_builder_end (builders[i]);
+          g_variant_builder_unref (builders[i]);
+          variant_set[i] = TRUE;
+        }
+    }
+
+  /* Check if all columns were set */
+  last_unset_col = -1;
+  for (i = 0; i < n_cols; i++)
+    {
+      if (!variant_set[i])
+        {
+          /* Create empty a{sv} if needed */
+          if (g_variant_type_is_subtype_of (G_VARIANT_TYPE (schema[i]),
+                                            G_VARIANT_TYPE_VARDICT))
+            {
+              GVariantBuilder builder;
+              g_variant_builder_init (&builder, G_VARIANT_TYPE (schema[i]));
+              out_row_members[i] = g_variant_builder_end (&builder);
+              variant_set[i] = TRUE;
+            }
+          else
+            {
+              last_unset_col = i;
+            }
+        }
+    }
+
+  if (last_unset_col >= 0)
+    {
+      /* Be nice and unref the variants we created */
+      for (i = 0; i < n_cols; i++)
+        {
+          if (variant_set[i])
+            {
+              g_variant_unref (g_variant_ref_sink (out_row_members[i]));
+              out_row_members[i] = NULL;
+            }
+        }
+
+      const gchar **names = dee_model_get_column_names (self, NULL);
+      g_critical ("Unable to build row: Column %d '%s' wasn't set",
+                  last_unset_col, names ? names[last_unset_col] : "unnamed");
+      return NULL;
+    }
+
+  return out_row_members;
+}
+
 /**
  * dee_model_append:
  * @self: a #DeeModel
@@ -595,25 +1214,18 @@ dee_model_append_valist (DeeModel *self,
   DeeModelIface *iface;
   DeeModelIter  *iter;
   GVariant     **row_members;
-  GVariant      *fast_row_members[DEFAULT_NUM_COLUMNS];
-  gsize          slice_size;
+  guint          num_columns;
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, &num_columns, return NULL);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
-  /* We use the slice allocator for this and not just g_new() since we
-   * can expect a lot of allocations of the same size */
-  slice_size = dee_model_get_n_columns (self) * sizeof (gpointer);
-  row_members = slice_size > sizeof (fast_row_members) ?
-    g_slice_alloc (slice_size) : fast_row_members;
+  row_members = g_alloca (num_columns * sizeof (gpointer));
 
   dee_model_build_row_valist (self, row_members, args);
 
   iter = (* iface->append_row) (self, row_members);
-  if (slice_size > sizeof (fast_row_members))
-      g_slice_free1 (slice_size, row_members);
   return iter;
 }
 
@@ -639,7 +1251,7 @@ dee_model_append_row (DeeModel  *self,
   DeeModelIface *iface;
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, NULL, return NULL);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
@@ -658,8 +1270,6 @@ dee_model_append_row (DeeModel  *self,
  * Creates and prepends a new row to the beginning of a #DeeModel, setting the
  * row values upon creation.
  *
- * Returns: (transfer none) (type Dee.ModelIter): A #DeeModelIter pointing to the new row
- *
  * Example:
  *
  * <informalexample><programlisting>
@@ -668,6 +1278,8 @@ dee_model_append_row (DeeModel  *self,
  *  dee_model_set_schema (model, "i", "s", NULL);
  *  dee_model_prepend (model, 10, "Rooney");
  * </programlisting></informalexample>
+ *
+ * Returns: (transfer none) (type Dee.ModelIter): A #DeeModelIter pointing to the new row
  */
 DeeModelIter*
 dee_model_prepend (DeeModel *self,
@@ -699,25 +1311,18 @@ dee_model_prepend_valist (DeeModel *self,
   DeeModelIface *iface;
   DeeModelIter  *iter;
   GVariant     **row_members;
-  GVariant      *fast_row_members[DEFAULT_NUM_COLUMNS];
-  gsize          slice_size;
+  guint          num_columns;
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, &num_columns, return NULL);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
-  /* We use the slice allocator for this and not just g_new() since we
-   * can expect a lot of allocations of the same size */
-  slice_size = dee_model_get_n_columns (self) * sizeof (gpointer);
-  row_members = slice_size > sizeof (fast_row_members) ?
-    g_slice_alloc (slice_size) : fast_row_members;
+  row_members = g_alloca (num_columns * sizeof (gpointer));
 
   dee_model_build_row_valist (self, row_members, args);
 
   iter = (* iface->prepend_row) (self, row_members);
-  if (slice_size > sizeof (fast_row_members))
-      g_slice_free1 (slice_size, row_members);
   return iter;
 }
 
@@ -744,7 +1349,7 @@ dee_model_prepend_row (DeeModel  *self,
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
 
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, NULL, return NULL);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
@@ -801,25 +1406,18 @@ dee_model_insert_valist (DeeModel *self,
   DeeModelIface *iface;
   DeeModelIter  *iter;
   GVariant     **row_members;
-  GVariant      *fast_row_members[DEFAULT_NUM_COLUMNS];
-  gsize          slice_size;
+  guint          num_columns;
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, &num_columns, return NULL);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
-  /* We use the slice allocator for this and not just g_new() since we
-   * can expect a lot of allocations of the same size */
-  slice_size = dee_model_get_n_columns (self) * sizeof (gpointer);
-  row_members = slice_size > sizeof (fast_row_members) ?
-    g_slice_alloc (slice_size) : fast_row_members;
+  row_members = g_alloca (num_columns * sizeof (gpointer));
 
   dee_model_build_row_valist (self, row_members, args);
 
   iter = (* iface->insert_row) (self, pos, row_members);
-  if (slice_size > sizeof (fast_row_members))
-      g_slice_free1 (slice_size, row_members);
   return iter;
 }
 
@@ -848,7 +1446,7 @@ dee_model_insert_row (DeeModel  *self,
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
 
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, NULL, return NULL);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
@@ -921,25 +1519,18 @@ dee_model_insert_before_valist (DeeModel     *self,
 {
   DeeModelIface  *iface;
   GVariant      **row_members;
-  GVariant       *fast_row_members[DEFAULT_NUM_COLUMNS];
-  gsize           slice_size;
+  guint           num_columns;
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, &num_columns, return NULL);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
-  /* We use the slice allocator for this and not just g_new() since we
-   * can expect a lot of allocations of the same size */
-  slice_size = dee_model_get_n_columns (self) * sizeof (gpointer);
-  row_members = slice_size > sizeof (fast_row_members) ?
-    g_slice_alloc (slice_size) : fast_row_members;
+  row_members = g_alloca (num_columns * sizeof (gpointer));
 
   dee_model_build_row_valist (self, row_members, args);
 
   iter = (* iface->insert_row_before) (self, iter, row_members);
-  if (slice_size > sizeof (fast_row_members))
-    g_slice_free1 (slice_size, row_members);
   return iter;
 }
 
@@ -968,7 +1559,7 @@ dee_model_insert_row_before (DeeModel      *self,
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
 
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, NULL, return NULL);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
@@ -1015,7 +1606,7 @@ dee_model_insert_row_sorted (DeeModel           *self,
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
 
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, NULL, return NULL);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
@@ -1081,29 +1672,22 @@ dee_model_insert_sorted (DeeModel           *self,
   DeeModelIface  *iface;
   DeeModelIter   *iter;
   GVariant      **row_members;
-  GVariant       *fast_row_members[DEFAULT_NUM_COLUMNS];
-  gsize           slice_size;
+  guint           num_columns;
   va_list         args;
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
 
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, &num_columns, return NULL);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
-  /* We use the slice allocator for this and not just g_new() since we
-   * can expect a lot of allocations of the same size */
-  slice_size = dee_model_get_n_columns (self) * sizeof (gpointer);
-  row_members = slice_size > sizeof (fast_row_members) ?
-    g_slice_alloc (slice_size) : fast_row_members;
+  row_members = g_alloca (num_columns * sizeof (gpointer));
 
   va_start (args, user_data);
   dee_model_build_row_valist (self, row_members, &args);
   va_end (args);
 
   iter = (* iface->insert_row_sorted) (self, row_members, cmp_func, user_data);
-  if (slice_size > sizeof (fast_row_members))
-    g_slice_free1 (slice_size, row_members);
   return iter;
 }
 
@@ -1128,7 +1712,7 @@ dee_model_insert_sorted (DeeModel           *self,
  * dee_model_insert_row_sorted() to insert rows in the model.
  *
  * Returns: (transfer none) (type Dee.ModelIter): If @out_was_found is set to
- *           %TRUE then a #DeeModelIter pointing to the first matching row.
+ *           %TRUE then a #DeeModelIter pointing to the last matching row.
  *           If it is %FALSE then the iter pointing to the row just after where
  *           @row_spec_would have been inserted.
  */
@@ -1143,7 +1727,7 @@ dee_model_find_row_sorted (DeeModel           *self,
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
 
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, NULL, return NULL);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
@@ -1176,7 +1760,7 @@ dee_model_find_row_sorted (DeeModel           *self,
  * to insert rows in the model.
  *
  * Returns: (transfer none) (type Dee.ModelIter): If @out_was_found is set to
- *           %TRUE then a #DeeModelIter pointing to the first matching row.
+ *           %TRUE then a #DeeModelIter pointing to the last matching row.
  *           If it is %FALSE then the iter pointing to the row just after where
  *           @row_spec_would have been inserted.
  */
@@ -1220,7 +1804,7 @@ dee_model_find_row_sorted_with_sizes (DeeModel                *self,
  * dee_model_insert_row_sorted() to insert rows in the model.
  *
  * Returns: (transfer none) (type Dee.ModelIter): If @out_was_found is set to
- *           %TRUE then a #DeeModelIter pointing to the first matching row.
+ *           %TRUE then a #DeeModelIter pointing to the last matching row.
  *           If it is %FALSE then the iter pointing to the row just after where
  *           @row_spec_would have been inserted.
  */
@@ -1234,21 +1818,16 @@ dee_model_find_sorted (DeeModel           *self,
   DeeModelIface  *iface;
   DeeModelIter   *iter;
   GVariant      **row_members;
-  GVariant       *fast_row_members[DEFAULT_NUM_COLUMNS];
-  gsize           slice_size;
+  guint           num_columns;
   va_list         args;
 
   g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
 
-  CHECK_SCHEMA (self, return NULL);
+  CHECK_SCHEMA (self, &num_columns, return NULL);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
-  /* We use the slice allocator for this and not just g_new() since we
-   * can expect a lot of allocations of the same size */
-  slice_size = dee_model_get_n_columns (self) * sizeof (gpointer);
-  row_members = slice_size > sizeof (fast_row_members) ?
-    g_slice_alloc (slice_size) : fast_row_members;
+  row_members = g_alloca (num_columns * sizeof (gpointer));
 
   va_start (args, out_was_found);
   dee_model_build_row_valist (self, row_members, &args);
@@ -1257,8 +1836,6 @@ dee_model_find_sorted (DeeModel           *self,
   iter = (* iface->find_row_sorted) (self, row_members, cmp_func,
                                      user_data, out_was_found);
 
-  if (slice_size > sizeof (fast_row_members))
-    g_slice_free1 (slice_size, row_members);
   return iter;
 }
 
@@ -1277,7 +1854,7 @@ dee_model_remove (DeeModel     *self,
 
   g_return_if_fail (DEE_IS_MODEL (self));
 
-  CHECK_SCHEMA (self, return);
+  CHECK_SCHEMA (self, NULL, return);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
@@ -1297,7 +1874,7 @@ dee_model_clear (DeeModel *self)
 
   g_return_if_fail (DEE_IS_MODEL (self));
 
-  CHECK_SCHEMA (self, return);
+  CHECK_SCHEMA (self, NULL, return);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
@@ -1363,21 +1940,17 @@ dee_model_set_valist (DeeModel       *self,
 {
   DeeModelIface  *iface;
   GVariant      **row_members;
-  GVariant       *fast_row_members[DEFAULT_NUM_COLUMNS];
-  gsize           slice_size;
+  guint           num_columns;
 
   g_return_if_fail (DEE_IS_MODEL (self));
 
   iface = DEE_MODEL_GET_IFACE (self);
-  slice_size = dee_model_get_n_columns (self) * sizeof(gpointer);
-  row_members = slice_size > sizeof (fast_row_members) ?
-    g_slice_alloc (slice_size) : fast_row_members;
+  num_columns = dee_model_get_n_columns (self);
+  row_members = g_alloca (num_columns * sizeof (gpointer));
 
   dee_model_build_row_valist (self, row_members, args);
 
   (* iface->set_row) (self, iter, row_members);
-  if (slice_size > sizeof (fast_row_members))
-    g_slice_free1 (slice_size, row_members);
 }
 
 /**
@@ -1405,7 +1978,7 @@ dee_model_set_value (DeeModel       *self,
 
   g_return_if_fail (DEE_IS_MODEL (self));
 
-  CHECK_SCHEMA (self, return);
+  CHECK_SCHEMA (self, NULL, return);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
@@ -1434,7 +2007,7 @@ dee_model_set_row (DeeModel       *self,
 
   g_return_if_fail (DEE_IS_MODEL (self));
 
-  CHECK_SCHEMA (self, return);
+  CHECK_SCHEMA (self, NULL, return);
 
   iface = DEE_MODEL_GET_IFACE (self);
 
@@ -1604,6 +2177,29 @@ dee_model_get_value (DeeModel     *self,
   return (* iface->get_value) (self, iter, column);
 }
 
+/**
+ * dee_model_get_value_by_name:
+ * @self: The #DeeModel to inspect
+ * @iter: a #DeeModelIter pointing to the row to inspect
+ * @column: column name to retrieve the value from
+ *
+ * Returns: (transfer full): A, guaranteed non-floating, reference to a
+ *          #GVariant containing the row data. Free with g_variant_unref().
+ */
+GVariant*
+dee_model_get_value_by_name (DeeModel     *self,
+                             DeeModelIter *iter,
+                             const gchar  *column_name)
+{
+  DeeModelIface *iface;
+
+  g_return_val_if_fail (DEE_IS_MODEL (self), NULL);
+
+  iface = DEE_MODEL_GET_IFACE (self);
+
+  return (* iface->get_value_by_name) (self, iter, column_name);
+}
+
 /**
  * dee_model_get_bool:
  * @self: a #DeeModel
