@@ -57,6 +57,7 @@ enum
   PROP_0,
   PROP_BACK_END,
   PROP_PROXY_SIGNALS,
+  PROP_INHERIT_SEQNUMS
 };
 
 /**
@@ -80,6 +81,8 @@ struct _DeeProxyModelPrivate
   gulong     row_added_handler;
   gulong     row_removed_handler;
   gulong     row_changed_handler;
+  gulong     changeset_started_handler;
+  gulong     changeset_finished_handler;
 };
 
 #define DEE_PROXY_MODEL_BACK_END(model) (DEE_PROXY_MODEL(model)->priv->back_end)
@@ -97,7 +100,28 @@ static const gchar* const* dee_proxy_model_get_schema (DeeModel *self,
 
 static const gchar*   dee_proxy_model_get_column_schema (DeeModel *self,
                                                          guint      column);
-                               
+
+static const gchar*   dee_proxy_model_get_field_schema  (DeeModel    *self,
+                                                         const gchar *field_name,
+                                                         guint       *out_column);
+
+static gint           dee_proxy_model_get_column_index (DeeModel    *self,
+                                                        const gchar *column_name);
+
+static void           dee_proxy_model_set_column_names (DeeModel     *self,
+                                                        const gchar **column_names,
+                                                        guint         num_columns);
+
+static const gchar**  dee_proxy_model_get_column_names (DeeModel *self,
+                                                        guint    *num_columns);
+
+static void           dee_proxy_model_register_vardict_schema (DeeModel   *self,
+                                                               guint       column,
+                                                               GHashTable *schema);
+
+static GHashTable*    dee_proxy_model_get_vardict_schema (DeeModel   *self,
+                                                          guint       column);
+
 static guint          dee_proxy_model_get_n_columns  (DeeModel *self);
 
 static guint          dee_proxy_model_get_n_rows     (DeeModel *self);
@@ -118,6 +142,17 @@ static DeeModelIter*  dee_proxy_model_insert_row_before (DeeModel      *self,
                                                          DeeModelIter  *iter,
                                                          GVariant     **row_members);
 
+static DeeModelIter*  dee_proxy_model_insert_row_sorted (DeeModel           *self,
+                                                         GVariant          **row_spec,
+                                                         DeeCompareRowFunc   cmp_func,
+                                                         gpointer            user_data);
+
+static DeeModelIter*  dee_proxy_model_find_row_sorted (DeeModel           *self,
+                                                       GVariant          **row_spec,
+                                                       DeeCompareRowFunc   cmp_func,
+                                                       gpointer            user_data,
+                                                       gboolean           *out_was_found);
+
 static void           dee_proxy_model_remove         (DeeModel     *self,
                                                       DeeModelIter *iter);
 
@@ -200,6 +235,10 @@ static void           dee_proxy_model_set_tag        (DeeModel       *self,
                                                       DeeModelTag    *tag,
                                                       gpointer        value);
 
+static void           dee_proxy_model_begin_changeset (DeeModel *self);
+
+static void           dee_proxy_model_end_changeset   (DeeModel *self);
+
 /*
  * Callbacks for relaying signals from the back end model
  */
@@ -212,6 +251,12 @@ static void           on_back_end_row_removed        (DeeProxyModel *self,
 static void           on_back_end_row_changed        (DeeProxyModel *self,
                                                       DeeModelIter  *iter);
 
+static void           on_back_end_changeset_started  (DeeProxyModel *self,
+                                                      DeeModel *model);
+
+static void           on_back_end_changeset_finished (DeeProxyModel *self,
+                                                      DeeModel *model);
+
 /*
  * Overrides for DeeSerializableModel
  */
@@ -237,6 +282,10 @@ dee_proxy_model_finalize (GObject *object)
         g_signal_handler_disconnect (priv->back_end, priv->row_removed_handler);
       if (priv->row_changed_handler != 0)
         g_signal_handler_disconnect (priv->back_end, priv->row_changed_handler);
+      if (priv->changeset_started_handler != 0)
+        g_signal_handler_disconnect (priv->back_end, priv->changeset_started_handler);
+      if (priv->changeset_finished_handler != 0)
+        g_signal_handler_disconnect (priv->back_end, priv->changeset_finished_handler);
 
       g_object_unref (priv->back_end);
     }
@@ -269,6 +318,16 @@ dee_proxy_model_constructed (GObject *object)
       priv->row_changed_handler =
         g_signal_connect_swapped (priv->back_end, "row-changed",
                                   G_CALLBACK (on_back_end_row_changed), object);
+
+      priv->changeset_started_handler =
+        g_signal_connect_swapped (priv->back_end, "changeset-started",
+                                  G_CALLBACK (on_back_end_changeset_started),
+                                  object);
+
+      priv->changeset_finished_handler =
+        g_signal_connect_swapped (priv->back_end, "changeset-finished",
+                                  G_CALLBACK (on_back_end_changeset_finished),
+                                  object);
     }
   
   /* GObjectClass has NULL 'constructed' member, but we add this check for
@@ -293,6 +352,9 @@ dee_proxy_model_set_property (GObject       *object,
     case PROP_PROXY_SIGNALS:
       priv->proxy_signals = g_value_get_boolean (value);
       break;
+    case PROP_INHERIT_SEQNUMS:
+      priv->inherit_seqnums = g_value_get_boolean (value);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, id, pspec);
       break;
@@ -313,6 +375,9 @@ dee_proxy_model_get_property (GObject     *object,
     case PROP_PROXY_SIGNALS:
       g_value_set_boolean (value, DEE_PROXY_MODEL (object)->priv->proxy_signals);
       break;
+    case PROP_INHERIT_SEQNUMS:
+      g_value_set_boolean (value, DEE_PROXY_MODEL (object)->priv->inherit_seqnums);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, id, pspec);
       break;
@@ -360,6 +425,22 @@ dee_proxy_model_class_init (DeeProxyModelClass *klass)
                                 G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
                                 | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (obj_class, PROP_PROXY_SIGNALS, pspec);
+
+  /**
+   * DeeProxyModel:inherit-seqnums:
+   *
+   * Boolean property defining whether sequence numbers will be inherited
+   * from the back end model.
+   * You will most likely want to set this property to false
+   * if the implementation manipulates with the rows in the model and keep
+   * track of seqnums yourself.
+   **/
+  pspec = g_param_spec_boolean ("inherit-seqnums", "Inherit seqnums",
+                                "Whether or not to inherit seqnums",
+                                TRUE,
+                                G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
+                                | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (obj_class, PROP_INHERIT_SEQNUMS, pspec);
   
   /* Add private data */
   g_type_class_add_private (obj_class, sizeof (DeeProxyModelPrivate));
@@ -368,39 +449,50 @@ dee_proxy_model_class_init (DeeProxyModelClass *klass)
 static void
 dee_proxy_model_model_iface_init (DeeModelIface *iface)
 {
-  iface->set_schema_full      = dee_proxy_model_set_schema_full;
-  iface->get_schema           = dee_proxy_model_get_schema;
-  iface->get_column_schema    = dee_proxy_model_get_column_schema;
-  iface->get_n_columns        = dee_proxy_model_get_n_columns;
-  iface->get_n_rows           = dee_proxy_model_get_n_rows;
-  iface->clear                = dee_proxy_model_clear;
-  iface->prepend_row          = dee_proxy_model_prepend_row;
-  iface->append_row           = dee_proxy_model_append_row;
-  iface->insert_row           = dee_proxy_model_insert_row;
-  iface->insert_row_before    = dee_proxy_model_insert_row_before;
-  iface->remove               = dee_proxy_model_remove;
-  iface->set_value            = dee_proxy_model_set_value;
-  iface->set_row              = dee_proxy_model_set_row;
-  iface->get_value            = dee_proxy_model_get_value;
-  iface->get_first_iter       = dee_proxy_model_get_first_iter;
-  iface->get_last_iter        = dee_proxy_model_get_last_iter;
-  iface->get_iter_at_row      = dee_proxy_model_get_iter_at_row;
-  iface->get_bool             = dee_proxy_model_get_bool;
-  iface->get_uchar            = dee_proxy_model_get_uchar;
-  iface->get_int32            = dee_proxy_model_get_int32;
-  iface->get_uint32           = dee_proxy_model_get_uint32;
-  iface->get_int64            = dee_proxy_model_get_int64;
-  iface->get_uint64           = dee_proxy_model_get_uint64;
-  iface->get_double           = dee_proxy_model_get_double;
-  iface->get_string           = dee_proxy_model_get_string;
-  iface->next                 = dee_proxy_model_next;
-  iface->prev                 = dee_proxy_model_prev;
-  iface->is_first             = dee_proxy_model_is_first;
-  iface->is_last              = dee_proxy_model_is_last;
-  iface->get_position         = dee_proxy_model_get_position;
-  iface->register_tag         = dee_proxy_model_register_tag;
-  iface->get_tag              = dee_proxy_model_get_tag;
-  iface->set_tag              = dee_proxy_model_set_tag;
+  iface->set_schema_full       = dee_proxy_model_set_schema_full;
+  iface->get_schema            = dee_proxy_model_get_schema;
+  iface->get_column_schema     = dee_proxy_model_get_column_schema;
+  iface->get_column_index      = dee_proxy_model_get_column_index;
+  iface->set_column_names_full = dee_proxy_model_set_column_names;
+  iface->get_column_names      = dee_proxy_model_get_column_names;
+  iface->get_field_schema      = dee_proxy_model_get_field_schema;
+  iface->get_n_columns         = dee_proxy_model_get_n_columns;
+  iface->get_n_rows            = dee_proxy_model_get_n_rows;
+  iface->clear                 = dee_proxy_model_clear;
+  iface->prepend_row           = dee_proxy_model_prepend_row;
+  iface->append_row            = dee_proxy_model_append_row;
+  iface->insert_row            = dee_proxy_model_insert_row;
+  iface->insert_row_before     = dee_proxy_model_insert_row_before;
+  iface->insert_row_sorted     = dee_proxy_model_insert_row_sorted;
+  iface->find_row_sorted       = dee_proxy_model_find_row_sorted;
+  iface->remove                = dee_proxy_model_remove;
+  iface->set_value             = dee_proxy_model_set_value;
+  iface->set_row               = dee_proxy_model_set_row;
+  iface->get_value             = dee_proxy_model_get_value;
+  iface->get_first_iter        = dee_proxy_model_get_first_iter;
+  iface->get_last_iter         = dee_proxy_model_get_last_iter;
+  iface->get_iter_at_row       = dee_proxy_model_get_iter_at_row;
+  iface->get_bool              = dee_proxy_model_get_bool;
+  iface->get_uchar             = dee_proxy_model_get_uchar;
+  iface->get_int32             = dee_proxy_model_get_int32;
+  iface->get_uint32            = dee_proxy_model_get_uint32;
+  iface->get_int64             = dee_proxy_model_get_int64;
+  iface->get_uint64            = dee_proxy_model_get_uint64;
+  iface->get_double            = dee_proxy_model_get_double;
+  iface->get_string            = dee_proxy_model_get_string;
+  iface->next                  = dee_proxy_model_next;
+  iface->prev                  = dee_proxy_model_prev;
+  iface->is_first              = dee_proxy_model_is_first;
+  iface->is_last               = dee_proxy_model_is_last;
+  iface->get_position          = dee_proxy_model_get_position;
+  iface->register_tag          = dee_proxy_model_register_tag;
+  iface->get_tag               = dee_proxy_model_get_tag;
+  iface->set_tag               = dee_proxy_model_set_tag;
+  iface->begin_changeset       = dee_proxy_model_begin_changeset;
+  iface->end_changeset         = dee_proxy_model_end_changeset;
+
+  iface->register_vardict_schema = dee_proxy_model_register_vardict_schema;
+  iface->get_vardict_schema      = dee_proxy_model_get_vardict_schema;
 }
 
 static void
@@ -410,11 +502,13 @@ dee_proxy_model_init (DeeProxyModel *model)
 
   priv = model->priv = DEE_PROXY_MODEL_GET_PRIVATE (model);
   priv->back_end = NULL;
-  priv->inherit_seqnums = FALSE;
+  priv->inherit_seqnums = TRUE;
   
   priv->row_added_handler = 0;
   priv->row_removed_handler = 0;
   priv->row_changed_handler = 0;
+  priv->changeset_started_handler = 0;
+  priv->changeset_finished_handler = 0;
 }
 
 /*
@@ -449,6 +543,73 @@ dee_proxy_model_get_column_schema (DeeModel *self,
   return dee_model_get_column_schema (DEE_PROXY_MODEL_BACK_END (self), column);
 }
 
+static const gchar*
+dee_proxy_model_get_field_schema (DeeModel    *self,
+                                  const gchar *field_name,
+                                  guint       *out_column)
+{
+  g_return_val_if_fail (DEE_IS_PROXY_MODEL (self), NULL);
+
+  return dee_model_get_field_schema (DEE_PROXY_MODEL_BACK_END (self),
+                                     field_name,
+                                     out_column);
+}
+
+static gint
+dee_proxy_model_get_column_index (DeeModel    *self,
+                                  const gchar *column_name)
+{
+  DeeModelIface *iface;
+
+  g_return_val_if_fail (DEE_IS_PROXY_MODEL (self), -1);
+
+  iface = DEE_MODEL_GET_IFACE (DEE_PROXY_MODEL_BACK_END (self));
+
+  return (* iface->get_column_index) (DEE_PROXY_MODEL_BACK_END (self),
+                                      column_name);
+}
+
+static void
+dee_proxy_model_set_column_names (DeeModel     *self,
+                                  const gchar **column_names,
+                                  guint         num_columns)
+{
+  g_return_if_fail (DEE_IS_PROXY_MODEL (self));
+
+  dee_model_set_column_names_full (DEE_PROXY_MODEL_BACK_END (self),
+                                   column_names, num_columns);
+}
+
+static const gchar**
+dee_proxy_model_get_column_names (DeeModel *self,
+                                  guint    *num_columns)
+{
+  g_return_val_if_fail (DEE_IS_PROXY_MODEL (self), NULL);
+
+  return dee_model_get_column_names (DEE_PROXY_MODEL_BACK_END (self),
+                                     num_columns);
+}
+
+static void
+dee_proxy_model_register_vardict_schema (DeeModel   *self,
+                                         guint       column,
+                                         GHashTable *schema)
+{
+  g_return_if_fail (DEE_IS_PROXY_MODEL (self));
+
+  dee_model_register_vardict_schema (DEE_PROXY_MODEL_BACK_END (self),
+                                     column, schema);
+}
+
+static GHashTable*
+dee_proxy_model_get_vardict_schema (DeeModel   *self,
+                                    guint       column)
+{
+  g_return_val_if_fail (DEE_IS_PROXY_MODEL (self), NULL);
+
+  return dee_model_get_vardict_schema (DEE_PROXY_MODEL_BACK_END (self), column);
+}
+
 static guint
 dee_proxy_model_get_n_columns (DeeModel *self)
 {
@@ -513,6 +674,32 @@ dee_proxy_model_insert_row_before (DeeModel      *self,
                                       iter, row_members);
 }
 
+static DeeModelIter*
+dee_proxy_model_insert_row_sorted (DeeModel           *self,
+                                   GVariant          **row_spec,
+                                   DeeCompareRowFunc   cmp_func,
+                                   gpointer            user_data)
+{
+  g_return_val_if_fail (DEE_IS_PROXY_MODEL (self), NULL);
+
+  return dee_model_insert_row_sorted (DEE_PROXY_MODEL_BACK_END (self),
+                                      row_spec, cmp_func, user_data);
+}
+
+static DeeModelIter*
+dee_proxy_model_find_row_sorted (DeeModel           *self,
+                                 GVariant          **row_spec,
+                                 DeeCompareRowFunc   cmp_func,
+                                 gpointer            user_data,
+                                 gboolean           *out_was_found)
+{
+  g_return_val_if_fail (DEE_IS_PROXY_MODEL (self), NULL);
+
+  return dee_model_find_row_sorted (DEE_PROXY_MODEL_BACK_END (self),
+                                    row_spec, cmp_func, user_data,
+                                    out_was_found);
+}
+
 static void
 dee_proxy_model_remove (DeeModel     *self,
                         DeeModelIter *iter)
@@ -734,6 +921,22 @@ dee_proxy_model_set_tag (DeeModel       *self,
   return dee_model_set_tag (DEE_PROXY_MODEL_BACK_END (self), iter, tag, value);
 }
 
+static void
+dee_proxy_model_begin_changeset (DeeModel *self)
+{
+  g_return_if_fail (DEE_IS_PROXY_MODEL (self));
+
+  dee_model_begin_changeset (DEE_PROXY_MODEL_BACK_END (self));
+}
+
+static void
+dee_proxy_model_end_changeset (DeeModel *self)
+{
+  g_return_if_fail (DEE_IS_PROXY_MODEL (self));
+
+  dee_model_end_changeset (DEE_PROXY_MODEL_BACK_END (self));
+}
+
 /*
  * Relay signals from back end
  */
@@ -759,6 +962,20 @@ on_back_end_row_changed (DeeProxyModel *self,
   g_signal_emit_by_name (self, "row-changed", iter);
 }
 
+static void
+on_back_end_changeset_started (DeeProxyModel *self,
+                               DeeModel *model)
+{
+  g_signal_emit_by_name (self, "changeset-started");
+}
+
+static void
+on_back_end_changeset_finished (DeeProxyModel *self,
+                                DeeModel *model)
+{
+  g_signal_emit_by_name (self, "changeset-finished");
+}
+
 /*
  * Overrides for DeeSerializableModel
  */
@@ -774,7 +991,7 @@ dee_proxy_model_get_seqnum (DeeModel     *self)
   if (priv->inherit_seqnums)
     return dee_serializable_model_get_seqnum (priv->back_end);
   else
-    return SUPER_CLASS->get_seqnum (priv->back_end);
+    return SUPER_CLASS->get_seqnum (self);
 }
 
 static void
@@ -790,7 +1007,7 @@ dee_proxy_model_set_seqnum  (DeeModel     *self,
   if (priv->inherit_seqnums)
     dee_serializable_model_set_seqnum (priv->back_end, seqnum);
   else
-    return SUPER_CLASS->set_seqnum (priv->back_end, seqnum);
+    return SUPER_CLASS->set_seqnum (self, seqnum);
 }
 
 static guint64
@@ -805,5 +1022,5 @@ dee_proxy_model_inc_seqnum (DeeModel     *self)
   if (priv->inherit_seqnums)
     return dee_serializable_model_inc_seqnum (priv->back_end);
   else
-    return SUPER_CLASS->inc_seqnum (priv->back_end);
+    return SUPER_CLASS->inc_seqnum (self);
 }
