@@ -58,6 +58,9 @@ G_DEFINE_ABSTRACT_TYPE_WITH_CODE (DeeSerializableModel,
 #define DEE_SERIALIZABLE_MODEL_GET_PRIVATE(obj) \
   (G_TYPE_INSTANCE_GET_PRIVATE(obj, DEE_TYPE_SERIALIZABLE_MODEL, DeeSerializableModelPrivate))
 
+#define MODEL_VARIANT_TYPE_1_0 G_VARIANT_TYPE ("(asaav(tt))")
+#define MODEL_VARIANT_TYPE     G_VARIANT_TYPE ("(asaav(tt)a{sv})")
+
 /**
  * DeeSerializableModelPrivate:
  *
@@ -71,7 +74,19 @@ struct _DeeSerializableModelPrivate
   /* Column type info */
   guint       n_columns;
   gchar     **column_schemas; // NULL terminated
-  
+  gchar     **column_names;   // NULL terminated
+  guint32    *column_name_hashes;
+  GHashTable *field_schemas;
+  gboolean    inside_changeset;
+};
+
+typedef struct _FieldSchemaInfo FieldSchemaInfo;
+
+struct _FieldSchemaInfo
+{
+  gint   ref_count;     // thread unsafe refcount
+  gchar *schema;
+  guint  column;
 };
 
 /*
@@ -91,12 +106,33 @@ static const gchar* const*   dee_serializable_model_get_schema (DeeModel *self,
                                                              guint    *num_columns);
 
 static const gchar*   dee_serializable_model_get_column_schema (DeeModel *self,
-                                                             guint     column);
+                                                                guint     column);
+
+static const gchar*   dee_serializable_model_get_field_schema  (DeeModel    *self,
+                                                                const gchar *field_name,
+                                                                guint       *out_column);
+
+static gint           dee_serializable_model_get_column_index  (DeeModel    *self,
+                                                                const gchar *column_name);
 
 static void           dee_serializable_model_set_schema_full  (DeeModel           *self,
                                                             const gchar* const *column_schemas,
                                                             guint               num_columns);
 
+static void           dee_serializable_model_set_column_names_full (DeeModel     *self,
+                                                                    const gchar **column_names,
+                                                                    guint         num_columns);
+
+static const gchar**  dee_serializable_model_get_column_names (DeeModel *self,
+                                                               guint    *num_columns);
+
+static void           dee_serializable_model_register_vardict_schema (DeeModel     *self,
+                                                                      guint         column,
+                                                                      GHashTable   *schema);
+
+static GHashTable*    dee_serializable_model_get_vardict_schema (DeeModel *self,
+                                                                 guint     column);
+
 static guint          dee_serializable_model_get_n_columns  (DeeModel *self);
 
 static guint          dee_serializable_model_get_n_rows     (DeeModel *self);
@@ -144,6 +180,10 @@ static GVariant*     dee_serializable_model_get_value      (DeeModel     *self,
                                                           DeeModelIter *iter,
                                                           guint         column);
 
+static GVariant*     dee_serializable_model_get_value_by_name (DeeModel     *self,
+                                                               DeeModelIter *iter,
+                                                               const gchar  *column_name);
+
 static GVariant**    dee_serializable_model_get_row         (DeeModel    *self,
                                                              DeeModelIter *iter,
                                                              GVariant    **out_row_members);
@@ -199,6 +239,50 @@ static gboolean       dee_serializable_model_is_first       (DeeModel     *self,
 static gboolean       dee_serializable_model_is_last        (DeeModel     *self,
                                                          DeeModelIter *iter);
 
+static void           dee_serializable_model_begin_changeset (DeeModel    *self);
+
+static void           dee_serializable_model_end_changeset   (DeeModel    *self);
+
+static guint sigid_changeset_started = 0;
+static guint sigid_changeset_finished = 0;
+
+/* FieldSchemaInfo methods */
+static FieldSchemaInfo*
+field_schema_info_new (const gchar *schema, guint column)
+{
+  FieldSchemaInfo *info;
+
+  info = g_slice_new (FieldSchemaInfo);
+  info->ref_count = 1;
+  info->schema = g_strdup (schema);
+  info->column = column;
+
+  return info;
+}
+
+static FieldSchemaInfo*
+field_schema_info_ref (FieldSchemaInfo *info)
+{
+  g_return_val_if_fail (info, NULL);
+
+  info->ref_count++;
+
+  return info;
+}
+
+static void
+field_schema_info_unref (FieldSchemaInfo *info)
+{
+  g_return_if_fail (info);
+  g_return_if_fail (info->ref_count > 0);
+
+  if (--info->ref_count <= 0)
+  {
+    g_free (info->schema);
+    g_slice_free (FieldSchemaInfo, info);
+  }
+}
+
 /* GObject Init */
 static void
 dee_serializable_model_finalize (GObject *object)
@@ -214,6 +298,24 @@ dee_serializable_model_finalize (GObject *object)
       priv->column_schemas = NULL;
     }
 
+  if (priv->column_names != NULL)
+    {
+      g_strfreev (priv->column_names);
+      priv->column_names = NULL;
+    }
+
+  if (priv->column_name_hashes != NULL)
+    {
+      g_free (priv->column_name_hashes);
+      priv->column_name_hashes = NULL;
+    }
+
+  if (priv->field_schemas != NULL)
+    {
+      g_hash_table_unref (priv->field_schemas);
+      priv->field_schemas = NULL;
+    }
+
   G_OBJECT_CLASS (dee_serializable_model_parent_class)->finalize (object);
 }
 
@@ -258,6 +360,9 @@ dee_serializable_model_class_init (DeeSerializableModelClass *klass)
   klass->set_seqnum = dee_serializable_model_set_seqnum_real;
   klass->inc_seqnum = dee_serializable_model_inc_seqnum_real;
 
+  sigid_changeset_started = g_signal_lookup ("changeset-started", DEE_TYPE_MODEL);
+  sigid_changeset_finished = g_signal_lookup ("changeset-finished", DEE_TYPE_MODEL);
+
   /* Add private data */
   g_type_class_add_private (obj_class, sizeof (DeeSerializableModelPrivate));
 }
@@ -401,15 +506,187 @@ dee_serializable_model_set_schema_full  (DeeModel           *self,
 #endif
 }
 
-static const gchar* const*
-dee_serializable_model_get_schema (DeeModel *self,
-                                guint    *n_columns)
+static void
+dee_serializable_model_set_column_names_full (DeeModel     *self,
+                                              const gchar **column_names,
+                                              guint         num_columns)
+{
+  DeeSerializableModelPrivate *priv;
+  guint i, j;
+  gboolean any_name_null;
+
+  g_return_if_fail (DEE_IS_SERIALIZABLE_MODEL (self));
+
+  priv = DEE_SERIALIZABLE_MODEL (self)->priv;
+
+  any_name_null = FALSE;
+  for (i = 0; i < num_columns; i++)
+    any_name_null |= column_names[i] == NULL;
+
+  if (num_columns < priv->n_columns || any_name_null)
+    {
+      g_critical ("All column names have to be set!");
+      return;
+    }
+
+  if (priv->column_names) g_strfreev (priv->column_names);
+  if (priv->column_name_hashes) g_free (priv->column_name_hashes);
+
+  priv->column_names = g_new0 (gchar*, priv->n_columns + 1);
+  priv->column_name_hashes = g_new0 (guint32, priv->n_columns);
+
+  for (i = 0; i < num_columns; i++)
+    {
+      priv->column_names[i] = g_strdup (column_names[i]);
+
+      // hash for fast compares
+      priv->column_name_hashes[i] = column_names[i] != NULL ?
+        g_str_hash (column_names[i]) : 0;
+    }
+
+  // check for uniqueness
+  for (i = 0; i < num_columns; i++)
+    {
+      for (j = i+1; j < num_columns; j++)
+        {
+          if (g_strcmp0 (priv->column_names[i], priv->column_names[j]) == 0)
+            {
+              g_warning ("Column names for columns %u and %u are the same!",
+                         i, j);
+            }
+        }
+    }
+}
+
+static const gchar**
+dee_serializable_model_get_column_names (DeeModel *self,
+                                         guint    *n_columns)
 {
   DeeSerializableModelPrivate *priv;
 
   g_return_val_if_fail (DEE_IS_SERIALIZABLE_MODEL (self), NULL);
 
-  priv = DEE_SERIALIZABLE_MODEL (self)->priv;
+  priv = ((DeeSerializableModel*) self)->priv;
+
+  if (n_columns != NULL)
+    *n_columns = priv->n_columns;
+
+  return (const gchar**) priv->column_names;
+}
+
+static void
+dee_serializable_model_register_vardict_schema (DeeModel     *self,
+                                                guint         column,
+                                                GHashTable   *schema)
+{
+  DeeSerializableModelPrivate *priv;
+  GHashTableIter iter;
+  gpointer       key, value;
+
+  g_return_if_fail (DEE_IS_SERIALIZABLE_MODEL (self));
+  g_return_if_fail (schema);
+
+  priv = ((DeeSerializableModel*) self)->priv;
+  g_return_if_fail (priv->column_schemas);
+  g_return_if_fail (column < priv->n_columns);
+  g_return_if_fail (g_variant_type_is_subtype_of (G_VARIANT_TYPE (priv->column_schemas[column]),
+                                                  G_VARIANT_TYPE_VARDICT));
+
+  if (priv->column_names == NULL || priv->column_names[column] == NULL)
+    {
+      g_critical ("Column name for column %u has to be set before calling "
+                  "dee_model_register_vardict_schema", column);
+      return;
+    }
+
+  if (priv->field_schemas == NULL)
+    priv->field_schemas = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                                 g_free,
+                                                 (GDestroyNotify) field_schema_info_unref);
+
+  g_hash_table_iter_init (&iter, schema);
+  while (g_hash_table_iter_next (&iter, &key, &value))
+    {
+      FieldSchemaInfo *info;
+      const gchar     *field_schema;
+      gchar *full_name;
+      guint  registered_column;
+
+      field_schema = dee_model_get_field_schema (self, key, &registered_column);
+      if (field_schema && registered_column != column)
+        {
+          g_warning ("Field '%s' is already registered for column %u! Please "
+                     "use fully qualified names to refer to it ('%s::%s' and "
+                     "'%s::%s')",
+                     (gchar*) key, registered_column,
+                     priv->column_names[registered_column], (gchar*) key,
+                     priv->column_names[column], (gchar*) key);
+        }
+      else if (field_schema && !g_str_equal (field_schema, value))
+        {
+          g_warning ("Field '%s' was already registered with schema '%s'! "
+                     "Overwriting with schema '%s'",
+                     (gchar*) key, field_schema, (gchar*) value);
+        }
+
+      info = field_schema_info_new (value, column);
+      g_hash_table_insert (priv->field_schemas, g_strdup (key), info);
+      full_name = g_strdup_printf ("%s::%s", priv->column_names[column],
+                                   (gchar*) key);
+      /* transfering ownership of full_name to hashtable */
+      g_hash_table_insert (priv->field_schemas, full_name,
+                           field_schema_info_ref (info));
+    }
+}
+
+static GHashTable*
+dee_serializable_model_get_vardict_schema (DeeModel     *self,
+                                           guint         column)
+{
+  DeeSerializableModelPrivate *priv;
+  GHashTable    *result;
+  GHashTableIter iter;
+  gpointer       key, value;
+
+  g_return_val_if_fail (DEE_IS_SERIALIZABLE_MODEL (self), NULL);
+
+  priv = ((DeeSerializableModel*) self)->priv;
+  g_return_val_if_fail (priv->column_schemas, NULL);
+  g_return_val_if_fail (column < priv->n_columns, NULL);
+  g_return_val_if_fail (g_variant_type_is_subtype_of (G_VARIANT_TYPE (priv->column_schemas[column]),
+                                                  G_VARIANT_TYPE_VARDICT), NULL);
+
+  if (priv->field_schemas == NULL) return NULL;
+
+  result = g_hash_table_new (g_str_hash, g_str_equal);
+  g_hash_table_iter_init (&iter, priv->field_schemas);
+  while (g_hash_table_iter_next (&iter, &key, &value))
+    {
+      gchar *field, *field_name;
+      FieldSchemaInfo *info;
+
+      field = (gchar*) key;
+      info = (FieldSchemaInfo*) value;
+      if (info->column != column) continue;
+
+      field_name = strstr (field, "::");
+      field_name = field_name != NULL ? field_name + 2 : field;
+
+      g_hash_table_insert (result, field_name, info->schema);
+    }
+
+  return result;
+}
+
+static const gchar* const*
+dee_serializable_model_get_schema (DeeModel *self,
+                                   guint    *n_columns)
+{
+  DeeSerializableModelPrivate *priv;
+
+  g_return_val_if_fail (DEE_IS_SERIALIZABLE_MODEL (self), NULL);
+
+  priv = ((DeeSerializableModel*) self)->priv;
 
   if (n_columns != NULL)
     *n_columns = priv->n_columns;
@@ -425,18 +702,67 @@ dee_serializable_model_get_column_schema (DeeModel *self,
   
   g_return_val_if_fail (DEE_IS_SERIALIZABLE_MODEL (self), NULL);
 
-  priv = DEE_SERIALIZABLE_MODEL (self)->priv;
+  priv = ((DeeSerializableModel*) self)->priv;
   g_return_val_if_fail (column < priv->n_columns, NULL);
 
   return priv->column_schemas[column];
 }
 
+static const gchar*
+dee_serializable_model_get_field_schema (DeeModel    *self,
+                                         const gchar *field_name,
+                                         guint       *out_column)
+{
+  DeeSerializableModelPrivate *priv;
+  FieldSchemaInfo *info;
+  
+  g_return_val_if_fail (DEE_IS_SERIALIZABLE_MODEL (self), NULL);
+  g_return_val_if_fail (field_name, NULL);
+
+  priv = ((DeeSerializableModel*) self)->priv;
+
+  if (priv->field_schemas == NULL) return NULL;
+
+  info = g_hash_table_lookup (priv->field_schemas, field_name);
+  if (info == NULL) return NULL;
+
+  if (out_column) *out_column = info->column;
+
+  return info->schema;
+}
+
+static gint
+dee_serializable_model_get_column_index (DeeModel    *self,
+                                         const gchar *column_name)
+{
+  gint                         i;
+  guint                        hash;
+  DeeSerializableModelPrivate *priv;
+  
+  g_return_val_if_fail (DEE_IS_SERIALIZABLE_MODEL (self), -1);
+
+  priv = ((DeeSerializableModel*) self)->priv;
+
+  if (priv->column_names == NULL || column_name == NULL) return -1;
+
+  hash = g_str_hash (column_name);
+
+  // FIXME: use a hashtable instead?
+  for (i = 0; i < priv->n_columns; i++)
+  {
+    if (priv->column_name_hashes[i] == hash &&
+        g_str_equal (priv->column_names[i], column_name)) return i;
+  }
+
+  return -1;
+}
+
 static guint
 dee_serializable_model_get_n_columns (DeeModel *self)
 {
   g_return_val_if_fail (DEE_IS_SERIALIZABLE_MODEL (self), 0);
 
-  return DEE_SERIALIZABLE_MODEL (self)->priv->n_columns;
+  return ((DeeSerializableModel*) self)->priv->n_columns;
 }
 
 static guint
@@ -534,13 +860,17 @@ dee_serializable_model_insert_row_sorted (DeeModel           *self,
                                           gpointer            user_data)
 {
   DeeModelIter *iter;
+  gboolean was_found;
 
   g_return_val_if_fail (DEE_IS_SERIALIZABLE_MODEL (self), NULL);
   g_return_val_if_fail (row_members != NULL, NULL);
   g_return_val_if_fail (cmp_func != NULL, NULL);
 
   iter = dee_model_find_row_sorted (self, row_members, cmp_func,
-                                    user_data, NULL);
+                                    user_data, &was_found);
+  if (was_found)
+    iter = dee_model_next (self, iter);
+
   return dee_model_insert_row_before (self, iter, row_members);
 }
 
@@ -551,14 +881,16 @@ dee_serializable_model_find_row_sorted   (DeeModel           *self,
                                           gpointer            user_data,
                                           gboolean           *out_was_found)
 {
-  DeeModelIter  *iter, *end;
+  DeeModelIter  *iter, *end, *last_matching;
   GVariant     **row_buf;
+  gint           cmp_result;
   guint          n_cols, i;
 
   g_return_val_if_fail (DEE_IS_SERIALIZABLE_MODEL (self), NULL);
   g_return_val_if_fail (row_spec != NULL, NULL);
   g_return_val_if_fail (cmp_func != NULL, NULL);
 
+  last_matching = NULL;
   if (out_was_found != NULL) *out_was_found = FALSE;
   n_cols = dee_model_get_n_columns (self);
 
@@ -570,18 +902,31 @@ dee_serializable_model_find_row_sorted   (DeeModel           *self,
   while (iter != end)
     {
       dee_model_get_row (self, iter, row_buf);
-      if (cmp_func(row_buf, row_spec, user_data) >= 0)
+      cmp_result = cmp_func (row_buf, row_spec, user_data);
+      /* we're returning last matching row to make ordering
+       * of insert_row_sorted stable and fast */
+      while (cmp_result == 0)
         {
+          last_matching = iter;
+          iter = dee_model_next (self, iter);
+          if (iter == end)
+            {
+              iter = last_matching;
+              break;
+            }
           for (i = 0; i < n_cols; i++) g_variant_unref (row_buf[i]);
-          if (out_was_found != NULL) *out_was_found = TRUE;
-          break;
+          dee_model_get_row (self, iter, row_buf);
+          cmp_result = cmp_func (row_buf, row_spec, user_data);
         }
 
       for (i = 0; i < n_cols; i++) g_variant_unref (row_buf[i]);
+      /* if we're past the matching row, the loop can be quit */
+      if (cmp_result >= 0) break;
       iter = dee_model_next (self, iter);
     }
 
-  return iter;
+  if (out_was_found != NULL && last_matching != NULL) *out_was_found = TRUE;
+  return last_matching ? last_matching : iter;
 }
 
 static void
@@ -618,6 +963,39 @@ dee_serializable_model_get_value (DeeModel     *self,
   return NULL;
 }
 
+static GVariant*
+dee_serializable_model_get_value_by_name (DeeModel     *self,
+                                          DeeModelIter *iter,
+                                          const gchar  *column_name)
+{
+  gint col_index;
+
+  g_return_val_if_fail (DEE_IS_SERIALIZABLE_MODEL (self), NULL);
+
+  col_index = dee_model_get_column_index (self, column_name);
+
+  if (col_index >= 0)
+    {
+      return dee_model_get_value (self, iter, col_index);
+    }
+  else if (dee_model_get_field_schema (self, column_name, (guint*) &col_index))
+    {
+      GVariant    *dict, *result;
+      const gchar *key_name;
+
+      dict = dee_model_get_value (self, iter, col_index);
+      // handle full "column::field" name
+      key_name = strstr(column_name, "::");
+      key_name = key_name != NULL ? key_name + 2 : column_name;
+      result = g_variant_lookup_value (dict, key_name, NULL);
+      g_variant_unref (dict);
+
+      return result;
+    }
+
+  return NULL;
+}
+
 static GVariant**
 dee_serializable_model_get_row (DeeModel      *self,
                                 DeeModelIter  *iter,
@@ -961,6 +1339,44 @@ dee_serializable_model_get_position (DeeModel     *self,
     }
 }
 
+static void
+dee_serializable_model_begin_changeset (DeeModel *self)
+{
+  DeeSerializableModelPrivate  *priv;
+  
+  priv = DEE_SERIALIZABLE_MODEL (self)->priv;
+
+  if (!priv->inside_changeset)
+    {
+      priv->inside_changeset = TRUE;
+      g_signal_emit (self, sigid_changeset_started, 0);
+    }
+  else
+    {
+      g_warning ("Ignored call to dee_model_begin_changeset, finish "
+                 "the current changeset using dee_model_end_changeset first");
+    }
+}
+
+static void
+dee_serializable_model_end_changeset (DeeModel *self)
+{
+  DeeSerializableModelPrivate  *priv;
+  
+  priv = DEE_SERIALIZABLE_MODEL (self)->priv;
+
+  if (priv->inside_changeset)
+    {
+      priv->inside_changeset = FALSE;
+      g_signal_emit (self, sigid_changeset_finished, 0);
+    }
+  else
+    {
+      g_warning ("Ignored call to dee_model_end_changeset, "
+                 "dee_model_begin_changeset has to be called first");
+    }
+}
+
 static DeeModelTag*
 dee_serializable_model_register_tag    (DeeModel       *self,
                                         GDestroyNotify  tag_destroy)
@@ -992,11 +1408,13 @@ static GVariant*
 dee_serializable_model_serialize (DeeSerializable *self)
 {
   DeeModel               *_self;
-  GVariantBuilder         aav, clone;
-  GVariant               *val, *tt, *schema;
+  GVariantBuilder         aav, clone, fields, vardict;
+  GVariant               *val, *tt, *schema, *col_names;
   DeeModelIter           *iter;
-  guint                   i, j, n_columns;
+  guint                   i, j, n_rows, n_columns;
   guint64                 last_seqnum;
+  const gchar* const     *column_schemas;
+  const gchar           **column_names;
 
   g_return_val_if_fail (DEE_IS_SERIALIZABLE_MODEL (self), FALSE);
 
@@ -1025,18 +1443,54 @@ dee_serializable_model_serialize (DeeSerializable *self)
       i++;
     }
 
+  n_rows = i;
+
   /* Collect the schema */
-  schema = g_variant_new_strv (dee_model_get_schema(_self, NULL), -1);
+  column_schemas = dee_model_get_schema(_self, NULL);
+  schema = g_variant_new_strv (column_schemas, -1);
+
+  /* Collect the column names */
+  column_names = dee_model_get_column_names (_self, NULL);
+  col_names = g_variant_new_strv (column_names,
+                                  column_names != NULL ? n_columns : 0);
+
+  g_variant_builder_init (&fields, G_VARIANT_TYPE ("a(uss)"));
+  for (i = 0; i < n_columns; i++)
+    {
+      GHashTable *field_schemas;
+      GHashTableIter ht_iter;
+      gpointer       key, value;
+
+      if (!g_variant_type_is_subtype_of (G_VARIANT_TYPE (column_schemas[i]),
+                                         G_VARIANT_TYPE_VARDICT))
+        continue;
+
+      field_schemas = dee_model_get_vardict_schema (_self, i);
+      if (field_schemas == NULL) continue;
+      g_hash_table_iter_init (&ht_iter, field_schemas);
+      while (g_hash_table_iter_next (&ht_iter, &key, &value))
+        {
+          g_variant_builder_add (&fields, "(uss)", i, key, value);
+        }
+
+      g_hash_table_unref (field_schemas);
+    }
 
   /* Collect the seqnum */
   last_seqnum = dee_serializable_model_get_seqnum (_self);
-  tt = g_variant_new ("(tt)", last_seqnum - i, last_seqnum);
+  tt = g_variant_new ("(tt)", last_seqnum - n_rows, last_seqnum);
+
+  /* Put all extra properties in a vardict */
+  g_variant_builder_init (&vardict, G_VARIANT_TYPE ("a{sv}"));
+  g_variant_builder_add (&vardict, "{sv}", "column-names", col_names);
+  g_variant_builder_add (&vardict, "{sv}", "fields", g_variant_builder_end (&fields));
 
   /* Build the final clone */
-  g_variant_builder_init (&clone, G_VARIANT_TYPE ("(asaav(tt))"));
+  g_variant_builder_init (&clone, MODEL_VARIANT_TYPE);
   g_variant_builder_add_value (&clone, schema);
   g_variant_builder_add_value (&clone, g_variant_builder_end (&aav));
   g_variant_builder_add_value (&clone, tt);
+  g_variant_builder_add_value (&clone, g_variant_builder_end (&vardict));
 
   trace_object (self, "Serialized with %i rows", dee_model_get_n_rows (_self));
 
@@ -1047,11 +1501,12 @@ static GObject*
 dee_serializable_model_parse_serialized (GVariant *data)
 {
   DeeModel      *model;
-  GVariant      *schemav, *seqnumsv, *col;
-  GVariantIter  *row_iter, *col_iter;
+  GVariant      *seqnumsv, *col, *vardict;
+  GVariantIter  *row_iter, *col_iter, *vardict_schema_iter;
   GVariant     **row;
   const gchar  **schemas;
-  gsize          n_cols, i, j;
+  const gchar  **column_names;
+  gsize          n_cols, i, j, tuple_items;
   guint64        seqnum_start, seqnum_end;
   static GType   default_model_type = G_TYPE_INVALID;
 
@@ -1065,14 +1520,81 @@ dee_serializable_model_parse_serialized (GVariant *data)
         }
     }
 
-  g_variant_get (data, "(@asaav@(tt))", &schemav, &row_iter, &seqnumsv);
-  schemas = g_variant_get_strv (schemav, &n_cols);
+  tuple_items = g_variant_n_children (data);
+  /* We support schema used by both dee 1.0 and 1.2 */
+  if (tuple_items == 3) /* "(asaav(tt))" */
+    {
+      g_variant_get (data, "(^a&saav@(tt))", &schemas, &row_iter, &seqnumsv);
+      vardict = NULL;
+    }
+  else if (tuple_items == 4) /* "(asaav(tt)a{sv})" */
+    {
+      g_variant_get (data, "(^a&saav@(tt)@a{sv})", &schemas, &row_iter,
+                     &seqnumsv, &vardict);
+
+      if (!g_variant_lookup (vardict, "column-names", "^a&s", &column_names))
+        column_names = NULL;
+      if (!g_variant_lookup (vardict, "fields", "a(uss)", &vardict_schema_iter))
+        vardict_schema_iter = NULL;
+    }
+  else
+    {
+      g_critical ("Unable to deserialize model: Unrecognized schema");
+      return NULL;
+    }
+
+  n_cols = g_strv_length ((gchar**) schemas);
   g_variant_get (seqnumsv, "(tt)", &seqnum_start, &seqnum_end);
 
   model = DEE_MODEL (g_object_new (default_model_type, NULL));
   dee_model_set_schema_full (model, schemas, n_cols);
   dee_serializable_model_set_seqnum (model, seqnum_start);
 
+  if (vardict)
+    {
+      if (column_names && g_strv_length ((gchar**) column_names) == n_cols)
+        {
+          dee_model_set_column_names_full (model, column_names, n_cols);
+        }
+
+      if (vardict_schema_iter != NULL)
+        {
+          GHashTable **vardict_schemas;
+          gchar *field_name, *field_schema;
+          guint column_index;
+
+          vardict_schemas = g_alloca (n_cols * sizeof (GHashTable*));
+          memset (vardict_schemas, 0, n_cols * sizeof (GHashTable*));
+
+          while (g_variant_iter_next (vardict_schema_iter, "(uss)",
+                                      &column_index, &field_name, &field_schema))
+            {
+              if (vardict_schemas[column_index] == NULL)
+                {
+                  vardict_schemas[column_index] = g_hash_table_new_full (
+                      g_str_hash, g_str_equal, g_free, g_free);
+                }
+
+              // using g_variant_iter_next, so we own field_name & schema
+              g_hash_table_insert (vardict_schemas[column_index],
+                                   field_name, field_schema);
+            }
+
+          for (column_index = 0; column_index < n_cols; column_index++)
+            {
+              if (vardict_schemas[column_index] == NULL) continue;
+              dee_model_register_vardict_schema (model, column_index,
+                                                 vardict_schemas[column_index]);
+              g_hash_table_unref (vardict_schemas[column_index]);
+            }
+
+          g_variant_iter_free (vardict_schema_iter);
+        }
+
+      g_free (column_names);
+      g_variant_unref (vardict);
+    }
+
   /* Note: The 'row' variable is stack allocated. No need to free it */
   row = g_alloca (n_cols * sizeof (GVariant *));
 
@@ -1091,9 +1613,9 @@ dee_serializable_model_parse_serialized (GVariant *data)
         }
 
       j = 0;
-      while (g_variant_iter_loop (col_iter, "v", &col))
+      while (g_variant_iter_next (col_iter, "v", &col))
         {
-          row[j] = g_variant_ref (col);
+          row[j] = col; // transfering variant reference to row[j]
           j++;
         }
 
@@ -1108,10 +1630,8 @@ dee_serializable_model_parse_serialized (GVariant *data)
       g_variant_iter_free (col_iter);
     }
   g_variant_iter_free (row_iter);
-
-  g_variant_unref (schemav);
-  g_variant_unref (seqnumsv);
   g_free (schemas);
+  g_variant_unref (seqnumsv);
 
   return (GObject *) model;
 }
@@ -1119,42 +1639,54 @@ dee_serializable_model_parse_serialized (GVariant *data)
 static void
 dee_serializable_model_model_iface_init (DeeModelIface *iface)
 {
-  iface->set_schema_full      = dee_serializable_model_set_schema_full;
-  iface->get_schema           = dee_serializable_model_get_schema;
-  iface->get_column_schema    = dee_serializable_model_get_column_schema;
-  iface->get_n_columns        = dee_serializable_model_get_n_columns;
-  iface->get_n_rows           = dee_serializable_model_get_n_rows;
-  iface->append_row           = dee_serializable_model_append_row;
-  iface->prepend_row          = dee_serializable_model_prepend_row;
-  iface->insert_row           = dee_serializable_model_insert_row;
-  iface->insert_row_before    = dee_serializable_model_insert_row_before;
-  iface->insert_row_sorted    = dee_serializable_model_insert_row_sorted;
-  iface->find_row_sorted      = dee_serializable_model_find_row_sorted;
-  iface->remove               = dee_serializable_model_remove;
-  iface->clear                = dee_serializable_model_clear;
-  iface->set_value            = dee_serializable_model_set_value;
-  iface->set_row              = dee_serializable_model_set_row;
-  iface->get_value            = dee_serializable_model_get_value;
-  iface->get_row              = dee_serializable_model_get_row;
-  iface->get_first_iter       = dee_serializable_model_get_first_iter;
-  iface->get_last_iter        = dee_serializable_model_get_last_iter;
-  iface->get_iter_at_row      = dee_serializable_model_get_iter_at_row;
-  iface->get_bool             = dee_serializable_model_get_bool;
-  iface->get_uchar            = dee_serializable_model_get_uchar;
-  iface->get_int32            = dee_serializable_model_get_int32;
-  iface->get_uint32           = dee_serializable_model_get_uint32;
-  iface->get_int64            = dee_serializable_model_get_int64;
-  iface->get_uint64           = dee_serializable_model_get_uint64;
-  iface->get_double           = dee_serializable_model_get_double;
-  iface->get_string           = dee_serializable_model_get_string;
-  iface->next                 = dee_serializable_model_next;
-  iface->prev                 = dee_serializable_model_prev;
-  iface->is_first             = dee_serializable_model_is_first;
-  iface->is_last              = dee_serializable_model_is_last;
-  iface->get_position         = dee_serializable_model_get_position;
-  iface->register_tag         = dee_serializable_model_register_tag;
-  iface->get_tag              = dee_serializable_model_get_tag;
-  iface->set_tag              = dee_serializable_model_set_tag;
+  iface->set_schema_full       = dee_serializable_model_set_schema_full;
+  iface->get_schema            = dee_serializable_model_get_schema;
+  iface->get_column_schema     = dee_serializable_model_get_column_schema;
+  iface->get_field_schema      = dee_serializable_model_get_field_schema;
+  iface->get_column_index      = dee_serializable_model_get_column_index;
+  iface->set_column_names_full = dee_serializable_model_set_column_names_full;
+  iface->get_column_names      = dee_serializable_model_get_column_names;
+  iface->get_n_columns         = dee_serializable_model_get_n_columns;
+  iface->get_n_rows            = dee_serializable_model_get_n_rows;
+  iface->append_row            = dee_serializable_model_append_row;
+  iface->prepend_row           = dee_serializable_model_prepend_row;
+  iface->insert_row            = dee_serializable_model_insert_row;
+  iface->insert_row_before     = dee_serializable_model_insert_row_before;
+  iface->insert_row_sorted     = dee_serializable_model_insert_row_sorted;
+  iface->find_row_sorted       = dee_serializable_model_find_row_sorted;
+  iface->remove                = dee_serializable_model_remove;
+  iface->clear                 = dee_serializable_model_clear;
+  iface->set_value             = dee_serializable_model_set_value;
+  iface->set_row               = dee_serializable_model_set_row;
+  iface->get_value             = dee_serializable_model_get_value;
+  iface->get_value_by_name     = dee_serializable_model_get_value_by_name;
+  iface->get_row               = dee_serializable_model_get_row;
+  iface->get_first_iter        = dee_serializable_model_get_first_iter;
+  iface->get_last_iter         = dee_serializable_model_get_last_iter;
+  iface->get_iter_at_row       = dee_serializable_model_get_iter_at_row;
+  iface->get_bool              = dee_serializable_model_get_bool;
+  iface->get_uchar             = dee_serializable_model_get_uchar;
+  iface->get_int32             = dee_serializable_model_get_int32;
+  iface->get_uint32            = dee_serializable_model_get_uint32;
+  iface->get_int64             = dee_serializable_model_get_int64;
+  iface->get_uint64            = dee_serializable_model_get_uint64;
+  iface->get_double            = dee_serializable_model_get_double;
+  iface->get_string            = dee_serializable_model_get_string;
+  iface->next                  = dee_serializable_model_next;
+  iface->prev                  = dee_serializable_model_prev;
+  iface->is_first              = dee_serializable_model_is_first;
+  iface->is_last               = dee_serializable_model_is_last;
+  iface->get_position          = dee_serializable_model_get_position;
+  iface->register_tag          = dee_serializable_model_register_tag;
+  iface->get_tag               = dee_serializable_model_get_tag;
+  iface->set_tag               = dee_serializable_model_set_tag;
+  iface->register_vardict_schema =
+    dee_serializable_model_register_vardict_schema;
+  iface->get_vardict_schema =
+    dee_serializable_model_get_vardict_schema;
+
+  iface->begin_changeset       = dee_serializable_model_begin_changeset;
+  iface->end_changeset         = dee_serializable_model_end_changeset;
 }
 
 static void
@@ -1163,6 +1695,10 @@ dee_serializable_model_serializable_iface_init (DeeSerializableIface *iface)
   iface->serialize      = dee_serializable_model_serialize;
 
   dee_serializable_register_parser (DEE_TYPE_SERIALIZABLE_MODEL,
-                                    G_VARIANT_TYPE ("(asaav(tt))"),
+                                    MODEL_VARIANT_TYPE_1_0,
+                                    dee_serializable_model_parse_serialized);
+  dee_serializable_register_parser (DEE_TYPE_SERIALIZABLE_MODEL,
+                                    MODEL_VARIANT_TYPE,
                                     dee_serializable_model_parse_serialized);
 }
+
