@@ -466,6 +466,15 @@ static void           dee_transaction_set_tag        (DeeModel       *self,
                                                       DeeModelTag    *tag,
                                                       gpointer        value);
 
+static void           dee_transaction_set_tag        (DeeModel       *self,
+                                                      DeeModelIter   *iter,
+                                                      DeeModelTag    *tag,
+                                                      gpointer        value);
+
+static const gchar*   dee_transaction_get_field_schema (DeeModel    *self,
+                                                        const gchar *field_name,
+                                                        guint       *out_column);
+
 static void           on_target_modified             (DeeTransaction *self,
                                                       DeeModelIter  *iter);
 
@@ -522,12 +531,13 @@ dee_transaction_finalize (GObject *object)
   G_OBJECT_CLASS (dee_transaction_parent_class)->finalize (object);
 }
 
-/* GObject Post-Init. Properties has been set */
+/* GObject Post-Init. Properties have been set */
 static void
 dee_transaction_constructed (GObject *object)
 {
-  const gchar* const    *schema;
-  guint                  n_columns;
+  const gchar* const  *schema;
+  const gchar        **column_names;
+  guint                n_columns;
   DeeTransactionPrivate *priv = DEE_TRANSACTION (object)->priv;
   
   /* Chain up before we do too much, to make sure the parent class is ready */
@@ -546,6 +556,14 @@ dee_transaction_constructed (GObject *object)
   dee_model_set_schema_full (DEE_MODEL (object), schema, n_columns);
   priv->n_cols = n_columns;
 
+  /* Also adopt column names of target model */
+  column_names = dee_model_get_column_names (priv->target, &n_columns);
+  if (column_names)
+    {
+      dee_model_set_column_names_full (DEE_MODEL (object),
+                                       column_names, n_columns);
+    }
+
   /* Adopt seqnums of target model, if it has any */
   if (DEE_IS_SERIALIZABLE_MODEL (priv->target))
     {
@@ -638,6 +656,7 @@ dee_transaction_model_iface_init (DeeModelIface *iface)
 
   iface->set_schema_full      = piface->set_schema_full;
   iface->get_schema           = piface->get_schema;
+  iface->get_field_schema     = dee_transaction_get_field_schema;
   iface->get_column_schema    = piface->get_column_schema;
   iface->get_n_columns        = piface->get_n_columns;
   iface->get_n_rows           = piface->get_n_rows;
@@ -702,6 +721,23 @@ dee_transaction_init (DeeTransaction *model)
 // dee_transaction_prepend_row ()
 // dee_transaction_insert_row ()
 
+static const gchar*
+dee_transaction_get_field_schema (DeeModel    *self,
+                                  const gchar *field_name,
+                                  guint       *out_column)
+{
+  DeeModelIface *piface;
+  const gchar   *schema;
+
+  schema = dee_model_get_field_schema (DEE_TRANSACTION_TARGET (self),
+                                       field_name, out_column);
+  if (schema) return schema;
+
+  // maybe the names are registered just for the transaction instance?
+  piface = g_type_interface_peek_parent (DEE_MODEL_GET_IFACE (self));
+  return (* piface->get_field_schema) (self, field_name, out_column);
+}
+
 static DeeModelIter*
 dee_transaction_insert_row_before (DeeModel      *self,
                                    DeeModelIter  *iter,
