@@ -31,7 +31,9 @@ main (gint argc, gchar *argv[])
 {
   DeeModel     *model;
   
+#if !GLIB_CHECK_VERSION(2, 35, 1)
   g_type_init (); 
+#endif
 
   if (argc == 2)
     model = dee_shared_model_new (argv[1]);
