@@ -40,8 +40,9 @@ main (gint argc, gchar *argv[])
   DeeModelIter *iter;
   gint          num_added;
   
+#if !GLIB_CHECK_VERSION(2, 35, 1)
   g_type_init (); 
-
+#endif
   g_set_prgname ("model-helper");
 
   if (argc == 2)
