@@ -155,6 +155,16 @@ bench_seqmodel_setup (Benchmark *bench)
   bench->state = model;
 }
 
+static void
+bench_seqmodel_named_setup (Benchmark *bench)
+{
+  bench_seqmodel_setup (bench);
+
+  dee_model_set_column_names (bench->state,
+                              "string1", "string2", "string3",
+                              "count", "bool", NULL);
+}
+
 static void
 bench_filter_model_collator_setup (Benchmark *bench)
 {
@@ -276,6 +286,36 @@ bench_model_append_run (Benchmark *bench)
     }
 }
 
+static void
+bench_model_named_append_run (Benchmark *bench)
+{
+  DeeModel  *model;
+  GVariant **row_buf;
+  guint      n_cols, i;
+  guint      limit = 2500;
+
+  g_assert (DEE_IS_MODEL (bench->state));
+
+  model = DEE_MODEL (bench->state);
+  n_cols = dee_model_get_n_columns (model);
+  row_buf = g_new0 (GVariant*, n_cols + 1);
+  
+  for (i = 0; i < limit; i++)
+    {
+      gchar *random_string = g_strdup_printf ("%"G_GINT32_FORMAT,
+                                              g_test_rand_int ());
+      dee_model_build_named_row (model, row_buf,
+                                 "count", (guint32) g_test_rand_int (),
+                                 "string1", random_string,
+                                 "string2", "Hello world",
+                                 "string3", "!",
+                                 "bool", TRUE,
+                                 NULL);
+      dee_model_append_row (model, row_buf);
+      g_free (random_string);
+    }
+}
+
 static void
 bench_model_prepend_run (Benchmark *bench)
 {
@@ -458,6 +498,13 @@ Benchmark seqmodel_append = { "SequenceModel.append",
                               100,
                               NULL };
 
+Benchmark seqmodel_named_append = { "SequenceModel.append.named",
+                                    bench_seqmodel_named_setup,
+                                    bench_model_named_append_run,
+                                    bench_gobject_teardown,
+                                    100,
+                                    NULL };
+
 Benchmark seqmodel_prepend = { "SequenceModel.prepend",
                                 bench_seqmodel_setup,
                                 bench_model_prepend_run,
@@ -540,7 +587,9 @@ Benchmark tree_index_prefix_search = { "TreeIndex.prefix_search",
 gint
 main (gint argc, gchar *argv[])
 {
+#if !GLIB_CHECK_VERSION(2, 35, 1)
   g_type_init ();
+#endif
   g_test_init (&argc, &argv, NULL);
   
   /* Extract NULL terminated array of prefixes from arguments */
@@ -552,6 +601,7 @@ main (gint argc, gchar *argv[])
     }
 
   add_benchmark (&seqmodel_append);
+  add_benchmark (&seqmodel_named_append);
   add_benchmark (&seqmodel_prepend);
   add_benchmark (&seqmodel_sorted);
   add_benchmark (&seqmodel_read_string);
