@@ -819,6 +819,12 @@ test_multiple_models2 (Fixture *fix, gconstpointer data)
   gtx_yield_main_loop (200);
 }
 
+static void
+changeset_signal (DeeModel *model, gboolean *value)
+{
+    *value = TRUE;
+}
+
 static void
 test_remote_append (Fixture *fix, gconstpointer data)
 {
@@ -829,6 +835,13 @@ test_remote_append (Fixture *fix, gconstpointer data)
   g_assert (dee_shared_model_is_leader (DEE_SHARED_MODEL (fix->model)));
   g_assert_cmpuint (dee_model_get_n_rows (fix->model), ==, 0);
 
+  gboolean got_changeset_start = FALSE;
+  gboolean got_changeset_finish = FALSE;
+  g_signal_connect (fix->model, "changeset-started",
+                    G_CALLBACK (changeset_signal), &got_changeset_start);
+  g_signal_connect (fix->model, "changeset-finished",
+                    G_CALLBACK (changeset_signal), &got_changeset_finish);
+
   if (gtx_wait_for_command (TESTDIR,
                             MODEL_HELPER (append1, MODEL_NAME),
                             2000))
@@ -838,6 +851,8 @@ test_remote_append (Fixture *fix, gconstpointer data)
 
   /* There should be a new row in the model */
   g_assert_cmpuint (dee_model_get_n_rows (fix->model), ==, 1);
+  g_assert (got_changeset_start);
+  g_assert (got_changeset_finish);
 }
 
 static void
