@@ -28,8 +28,16 @@ typedef struct
 
 } FilterFixture;
 
-static void setup    (FilterFixture *fix, gconstpointer data);
-static void teardown (FilterFixture *fix, gconstpointer data);
+typedef struct
+{
+  gint first;
+  gint second;
+
+} TwoIntsTuple;
+
+static void setup       (FilterFixture *fix, gconstpointer data);
+static void setup_empty (FilterFixture *fix, gconstpointer data);
+static void teardown    (FilterFixture *fix, gconstpointer data);
 
 static void test_empty_orig                    (FilterFixture *fix,
                                                 gconstpointer  data);
@@ -39,6 +47,8 @@ static void test_discard_all                   (FilterFixture *fix,
                                                 gconstpointer  data);
 static void test_discard_all_append_notify     (FilterFixture *fix,
                                                 gconstpointer  data);
+static void test_change_backend                (FilterFixture *fix,
+                                                gconstpointer  data);
 static void test_collator_asc                  (FilterFixture *fix,
                                                 gconstpointer  data);
 static void test_collator_desc                  (FilterFixture *fix,
@@ -52,6 +62,9 @@ static void test_any                           (FilterFixture *fix,
 static void test_regex                         (FilterFixture *fix,
                                                 gconstpointer  data);
 
+static void test_changesets                    (FilterFixture *fix,
+                                                gconstpointer  data);
+
 void
 test_filter_model_create_suite (void)
 {
@@ -64,6 +77,8 @@ test_filter_model_create_suite (void)
               setup, test_discard_all, teardown);
   g_test_add (DOMAIN"/DiscardAllAppendNotify", FilterFixture, 0,
               setup, test_discard_all_append_notify, teardown);
+  g_test_add (DOMAIN"/ChangeBackend", FilterFixture, 0,
+              setup, test_change_backend, teardown);
   g_test_add (DOMAIN"/CollatorAscending", FilterFixture, 0,
               setup, test_collator_asc, teardown);
   g_test_add (DOMAIN"/CollatorDescending", FilterFixture, 0,
@@ -71,22 +86,35 @@ test_filter_model_create_suite (void)
   g_test_add (DOMAIN"/Key", FilterFixture, 0,
               setup, test_key, teardown);
   g_test_add (DOMAIN"/Any", FilterFixture, 0,
-                setup, test_any, teardown);
+              setup, test_any, teardown);
   g_test_add (DOMAIN"/Regex", FilterFixture, 0,
               setup, test_regex, teardown);
+  g_test_add (DOMAIN"/Changesets", FilterFixture, 0,
+              setup_empty, test_changesets, teardown);
+}
+
+static void
+setup_empty (FilterFixture *fix, gconstpointer data)
+{
+  fix->model = dee_sequence_model_new ();
+  dee_model_set_schema (fix->model, "i", "s", NULL);
+  
+  g_assert (DEE_IS_SEQUENCE_MODEL (fix->model));
+}
+
+static void
+add_3rows(DeeModel *model)
+{
+  dee_model_append (model, 0, "Zero");
+  dee_model_append (model, 1, "One");
+  dee_model_append (model, 2, "Two");
 }
 
 static void
 setup (FilterFixture *fix, gconstpointer data)
 {
-  fix->model = dee_sequence_model_new ();
-  dee_model_set_schema (fix->model, "i", "s", NULL);
-  
-  dee_model_append (fix->model, 0, "Zero");
-  dee_model_append (fix->model, 1, "One");
-  dee_model_append (fix->model, 2, "Two");
-
-  g_assert (DEE_IS_SEQUENCE_MODEL (fix->model));
+  setup_empty (fix, data);
+  add_3rows (fix->model);
 }
 
 static void
@@ -171,6 +199,8 @@ test_empty_orig (FilterFixture *fix, gconstpointer data)
   DeeModelIter *iter = dee_model_get_first_iter (m);
   g_assert (dee_model_is_first (m, iter));
   g_assert (dee_model_is_last (m, iter));
+
+  g_assert_cmpuint (0, ==, dee_serializable_model_get_seqnum (m));
 }
 
 /* A filter model that is a complete copy of the orig */
@@ -199,18 +229,23 @@ test_append_all (FilterFixture *fix, gconstpointer data)
   g_assert (!dee_model_is_last (m, iter));
   g_assert_cmpint (1, ==, dee_model_get_int32 (m, iter, 0));
   g_assert_cmpstr ("One", ==, dee_model_get_string (m, iter, 1));
-  
+  g_assert (dee_model_get_first_iter (m) == dee_model_prev (m, iter));
+
   iter = dee_model_next (m, iter);
   g_assert (!dee_model_is_first (m, iter));
   g_assert (!dee_model_is_last (m, iter));
   g_assert_cmpint (2, ==, dee_model_get_int32 (m, iter, 0));
   g_assert_cmpstr ("Two", ==, dee_model_get_string (m, iter, 1));
-  
+  g_assert (iter == dee_model_prev (m, dee_model_get_last_iter (m)));
+
   iter = dee_model_next (m, iter);
   g_assert (!dee_model_is_first (m, iter));
   g_assert (dee_model_is_last (m, iter));
   
   g_assert (dee_model_is_last (m, iter));
+
+  g_assert_cmpuint (dee_serializable_model_get_seqnum (fix->model), ==,
+                    dee_serializable_model_get_seqnum (m));
 }
 
 /* Test a filter that blocks everything */
@@ -236,7 +271,10 @@ test_discard_all (FilterFixture *fix, gconstpointer data)
   DeeModelIter *iter = dee_model_get_first_iter (m);
   g_assert (dee_model_is_first (m, iter));
   g_assert (dee_model_is_last (m, iter));
-  
+
+  /* Check that seqnum is still zero */
+  g_assert_cmpuint (0, ==, dee_serializable_model_get_seqnum (m));
+
   guint filter_add_count = 0;
   guint orig_add_count = 0;
   g_signal_connect (m, "row-added", G_CALLBACK (signal_counter), &filter_add_count);
@@ -250,6 +288,7 @@ test_discard_all (FilterFixture *fix, gconstpointer data)
   g_assert_cmpint (2, ==, orig_add_count);
   g_assert_cmpint (0, ==, dee_model_get_n_rows (m));
   g_assert_cmpint (5, ==, dee_model_get_n_rows (fix->model));
+  g_assert_cmpuint (0, ==, dee_serializable_model_get_seqnum (m));
   
   /* Now add stuff to the filtered model directly. This should work,
    * and should be written through to the orig model as well  */
@@ -258,6 +297,7 @@ test_discard_all (FilterFixture *fix, gconstpointer data)
   g_assert_cmpint (3, ==, orig_add_count);
   g_assert_cmpint (1, ==, dee_model_get_n_rows (m));
   g_assert_cmpint (6, ==, dee_model_get_n_rows (fix->model));
+  g_assert_cmpuint (1, ==, dee_serializable_model_get_seqnum (m));
   
   /* The first (and only) row of 'm' should be at offset 5 in fix->model */
   iter = dee_model_get_first_iter (m);
@@ -266,7 +306,28 @@ test_discard_all (FilterFixture *fix, gconstpointer data)
   g_assert_cmpint (27, ==, dee_model_get_int32 (fix->model, iter, 0));
   g_assert_cmpstr ("TwentySeven", ==, dee_model_get_string (m, iter, 1));
   g_assert_cmpstr ("TwentySeven", ==, dee_model_get_string (fix->model, iter, 1));
-  
+
+  /* And append two more rows to the filtered model, to ensure the order */  
+  dee_model_prepend (m, -1, "MinusOne");
+  dee_model_append (m, 39, "ThirtyNine");
+  g_assert_cmpint (3, ==, filter_add_count);
+  g_assert_cmpint (5, ==, orig_add_count);
+  g_assert_cmpint (3, ==, dee_model_get_n_rows (m));
+  g_assert_cmpint (8, ==, dee_model_get_n_rows (fix->model));
+  g_assert_cmpuint (3, ==, dee_serializable_model_get_seqnum (m));
+  iter = dee_model_prev (m, dee_model_get_last_iter (m));
+  g_assert_cmpint (39, ==, dee_model_get_int32 (m, iter, 0));
+  g_assert_cmpstr ("ThirtyNine", ==, dee_model_get_string (m, iter, 1));
+  iter = dee_model_prev (m, iter);
+  g_assert_cmpint (27, ==, dee_model_get_int32 (m, iter, 0));
+  g_assert_cmpstr ("TwentySeven", ==, dee_model_get_string (m, iter, 1));
+  iter = dee_model_prev (m, iter);
+  g_assert_cmpint (-1, ==, dee_model_get_int32 (m, iter, 0));
+  g_assert_cmpstr ("MinusOne", ==, dee_model_get_string (m, iter, 1));
+  iter = dee_model_prev (fix->model, dee_model_get_last_iter (fix->model));
+  g_assert_cmpint (39, ==, dee_model_get_int32 (fix->model, iter, 0));
+  g_assert_cmpstr ("ThirtyNine", ==, dee_model_get_string (fix->model, iter, 1));
+
   g_object_unref (m);
 }
 
@@ -283,6 +344,9 @@ test_discard_all_append_notify (FilterFixture *fix, gconstpointer data)
                   &filter);
 
   DeeModel *m = dee_filter_model_new (fix->model, &filter);
+
+  /* Nothing was added to the filter model, seqnum should be zero */
+  g_assert_cmpuint (0, ==, dee_serializable_model_get_seqnum (m));
   
   guint filter_add_count = 0;
   guint orig_add_count = 0;
@@ -297,6 +361,7 @@ test_discard_all_append_notify (FilterFixture *fix, gconstpointer data)
   g_assert_cmpint (2, ==, orig_add_count);
   g_assert_cmpint (2, ==, dee_model_get_n_rows (m));
   g_assert_cmpint (5, ==, dee_model_get_n_rows (fix->model));
+  g_assert_cmpuint (2, ==, dee_serializable_model_get_seqnum (m));
   
   /* The first row of 'm' should be at offset 3 in fix->model */
   DeeModelIter *iter = dee_model_get_first_iter (m);
@@ -321,6 +386,62 @@ test_discard_all_append_notify (FilterFixture *fix, gconstpointer data)
   g_object_unref (m);
 }
 
+static void
+test_change_backend (FilterFixture *fix, gconstpointer data)
+{
+  DeeFilter filter;
+
+  dee_filter_new (append_all_model_map,
+                  append_all_model_notify,
+                  fix,
+                  NULL,
+                  &filter);
+  
+  DeeModel *m = dee_filter_model_new (fix->model, &filter);
+  g_assert_cmpint (3, ==, dee_model_get_n_rows (fix->model));
+  
+  DeeModelIter *iter = dee_model_get_first_iter (m);
+  g_assert (dee_model_is_first (m, iter));
+  g_assert (!dee_model_is_last (m, iter));
+  g_assert_cmpint (0, ==, dee_model_get_int32 (m, iter, 0));
+  g_assert_cmpstr ("Zero", ==, dee_model_get_string (m, iter, 1));
+  g_assert_cmpuint (0, ==, dee_model_get_position (m, iter));
+
+  iter = dee_model_get_first_iter (fix->model);
+  dee_model_remove (fix->model, iter);
+
+  g_assert_cmpint (2, ==, dee_model_get_n_rows (fix->model));
+  g_assert_cmpint (2, ==, dee_model_get_n_rows (m));
+
+  iter = dee_model_get_first_iter (fix->model);
+  dee_model_remove (fix->model, iter);
+
+  g_assert_cmpint (1, ==, dee_model_get_n_rows (fix->model));
+  g_assert_cmpint (1, ==, dee_model_get_n_rows (m));
+
+  iter = dee_model_get_first_iter (m);
+  g_assert (dee_model_is_first (m, iter));
+  g_assert (!dee_model_is_last (m, iter));
+  g_assert_cmpint (2, ==, dee_model_get_int32 (m, iter, 0));
+  g_assert_cmpstr ("Two", ==, dee_model_get_string (m, iter, 1));
+
+  iter = dee_model_get_first_iter (fix->model);
+  dee_model_set (fix->model, iter, -1, "Minus one");
+
+  iter = dee_model_get_first_iter (m);
+  g_assert (dee_model_is_first (m, iter));
+  g_assert (!dee_model_is_last (m, iter));
+  g_assert_cmpint (-1, ==, dee_model_get_int32 (m, iter, 0));
+  g_assert_cmpstr ("Minus one", ==, dee_model_get_string (m, iter, 1));
+  g_assert_cmpuint (0, ==, dee_model_get_position (m, iter));
+
+  /* and finally remove the last row via the filter model */
+  dee_model_remove (m, iter);
+
+  g_assert_cmpint (0, ==, dee_model_get_n_rows (m));
+  g_assert_cmpint (0, ==, dee_model_get_n_rows (fix->model));
+}
+
 /* Test dee_filter_new_collator() ascending */
 static void
 test_collator_asc (FilterFixture *fix, gconstpointer data)
@@ -540,3 +661,88 @@ test_regex (FilterFixture *fix, gconstpointer data)
   _test_orig_ordering (fix, &filter);
   g_regex_unref (regex);
 }
+
+static void
+increment_first (TwoIntsTuple *tuple)
+{
+  tuple->first++;
+  // first always has to be incremented before second
+  g_assert (tuple->first > tuple->second);
+}
+
+static void
+increment_second (TwoIntsTuple *tuple)
+{
+  tuple->second++;
+  // second needs to be incremented after first
+  g_assert (tuple->second == tuple->first);
+}
+
+static void
+test_changesets (FilterFixture *fix, gconstpointer data)
+{
+  GRegex       *regex;
+  DeeFilter     filter;
+  DeeModel     *filter_m1;
+  DeeModel     *filter_m2;
+  DeeModel     *filter_m3;
+
+  regex = g_regex_new ("^..[eo]", 0, 0, NULL);
+  dee_filter_new_regex (1, regex, &filter);
+  filter_m1 = dee_filter_model_new (fix->model, &filter);
+  g_regex_unref (regex);
+
+  regex = g_regex_new ("^Z", 0, 0, NULL);
+  dee_filter_new_regex (1, regex, &filter);
+  filter_m2 = dee_filter_model_new (fix->model, &filter);
+  g_regex_unref (regex);
+
+  regex = g_regex_new ("^X", 0, 0, NULL);
+  dee_filter_new_regex (1, regex, &filter);
+  filter_m3 = dee_filter_model_new (fix->model, &filter);
+  g_regex_unref (regex);
+
+  g_assert_cmpuint (0, ==, dee_model_get_n_rows (fix->model));
+  g_assert_cmpuint (0, ==, dee_model_get_n_rows (filter_m1));
+  g_assert_cmpuint (0, ==, dee_model_get_n_rows (filter_m2));
+  g_assert_cmpuint (0, ==, dee_model_get_n_rows (filter_m3));
+
+  TwoIntsTuple tuple_m0 = { 0, 0 };
+  TwoIntsTuple tuple_m1 = { 0, 0 };
+  TwoIntsTuple tuple_m2 = { 0, 0 };
+  TwoIntsTuple tuple_m3 = { 0, 0 };
+
+  g_signal_connect_swapped (fix->model, "changeset-started",
+                            G_CALLBACK (increment_first), &tuple_m0);
+  g_signal_connect_swapped (fix->model, "changeset-finished",
+                            G_CALLBACK (increment_second), &tuple_m0);
+  g_signal_connect_swapped (filter_m1, "changeset-started",
+                            G_CALLBACK (increment_first), &tuple_m1);
+  g_signal_connect_swapped (filter_m1, "changeset-finished",
+                            G_CALLBACK (increment_second), &tuple_m1);
+  g_signal_connect_swapped (filter_m2, "changeset-started",
+                            G_CALLBACK (increment_first), &tuple_m2);
+  g_signal_connect_swapped (filter_m2, "changeset-finished",
+                            G_CALLBACK (increment_second), &tuple_m2);
+  g_signal_connect_swapped (filter_m3, "changeset-started",
+                            G_CALLBACK (increment_first), &tuple_m3);
+  g_signal_connect_swapped (filter_m3, "changeset-finished",
+                            G_CALLBACK (increment_second), &tuple_m3);
+
+  dee_model_begin_changeset (fix->model);
+  add_3rows (fix->model);
+  dee_model_end_changeset (fix->model);
+
+  g_assert_cmpuint (3, ==, dee_model_get_n_rows (fix->model));
+  g_assert_cmpuint (2, ==, dee_model_get_n_rows (filter_m1));
+  g_assert_cmpuint (1, ==, dee_model_get_n_rows (filter_m2));
+  g_assert_cmpuint (0, ==, dee_model_get_n_rows (filter_m3));
+  g_assert_cmpint (tuple_m0.first, ==, 1);
+  g_assert_cmpint (tuple_m0.second, ==, 1);
+  g_assert_cmpint (tuple_m1.first, ==, 1);
+  g_assert_cmpint (tuple_m1.second, ==, 1);
+  g_assert_cmpint (tuple_m2.first, ==, 1);
+  g_assert_cmpint (tuple_m2.second, ==, 1);
+  g_assert_cmpint (tuple_m3.first, ==, 1);
+  g_assert_cmpint (tuple_m3.second, ==, 1);
+}
